from rumboot.ops.base import base

# HINT: most exits are handled in xmodem class, since 
# this is also the place for incremental uploads
class termination(base):
    formats = {
        "panic" : "PANIC: {}",
        }

    def action(self, trigger, result):
        if trigger == "panic":
            return -1



from rumboot.chips.base import chipBase

class chipZynq(chipBase):
    name="zed"
    part="Xilinx Zynq series (Zed Board, Tang Hex, ...)"
    chip_id=255
    baudrate=115200
    chip_rev=1
    welcome='Zynq> '
    baudrate=115200
    memories = {}
    spl_address = 0x100
    hacks = {
        "skipsync"   : True, # Doesn't Send U\r\n\r\n at the start. All legacy stuff
        "silentRom"  : False, # Bootrom is totally silent
        "noxmodem"   : False,  # Chip lacks xmodem implementation
    }

class chipRPI4(chipBase):
    name="rpi4"
    part="BCM2711 (Raspberry Pi 4)"
    chip_id=255
    baudrate=115200
    chip_rev=2
    welcome='U-Boot> '
    baudrate=115200
    memories = {}


from rumboot.ops.base import base
import tqdm
import time

class basic_uploader(base):
    formats = {
        "first_upload"      : "boot: host: Hit '{}' for X-Modem upload",
        "first_upload_basis"  : "boot: host: Hit 'X' for xmodem upload",
        "upload_uboot": "Trying to boot from UART",
        "uboot_xmodem": "## Ready for binary (xmodem) download to {} at {} bps..."
        }

    def __init__(self, term):
        super().__init__(term)

    def sync(self, syncword, short = False):
        ser = self.term.ser
        if self.term.replay:
            return
        while True:
            ser.write(syncword.encode())
            if short:
                break
            while True:
                tmp1 = ser.read(1)
                tmp2 = ser.read(1)
                if tmp1 == b"C" and tmp2 == b"C": 
                    return
            break

    def action(self, trigger, result):
        if trigger != "upload_uboot" and self.term.xfer.how == "xmodem":
            self.sync("X")

        if not self.term.xfer.push(self.term.chip.spl_address):
            print("Upload failed")
            return 1
        return True

class smart_uploader(basic_uploader):
    formats = {
        "upload" : "UPLOAD to {:x}. 'X' for X-modem, 'E' for EDCL"
    }

    def action(self, trigger, result):
        if (self.term.xfer.how == "xmodem"):
            self.sync("X", True)

        if not self.term.xfer.push(result[0]):
            print("Upload failed")
            return 1

        if (self.term.xfer.how != "xmodem"):
            self.sync('E', True)
        return True

class smart_downloader(basic_uploader):
    formats = {
        "upload" : "DOWNLOAD: {:d} bytes from {:x} to {}. 'X' for X-modem, 'E' for EDCL"
    }

    def action(self, trigger, result):
        print(result)
        arg = result[2]
        if arg in self.term.plusargs:
            fl = self.term.plusargs[arg]
        else:
            fl = arg
        stream = open(fl, 'wb')

        if (self.term.xfer.how == "xmodem"):
            self.sync("X", True)

        self.term.xfer.recv(stream, result[1], result[0])

        if (self.term.xfer.how != "xmodem"):
            self.sync("E", True)

        pass


class runtime(basic_uploader):
    formats = {
        "runtime"           : "UPLOAD: {} to {:x}. 'X' for X-modem, 'E' for EDCL",
    }

    def action(self, trigger, result):
        arg = result[0]
        fl = self.term.plusargs[arg]
        stream = open(fl, 'rb')
        if (self.term.xfer.how == "xmodem"):
            self.sync("X", True)
        ret = self.term.xfer.send(stream, result[1], "Uploading")
        stream.close()  
        if (self.term.xfer.how != "xmodem"):
            self.sync('E', True)
        return ret



class incremental(basic_uploader):
    formats = {
        "incremental_upload": "boot: host: Back in rom, code {}",
    }

    def action(self, trigger, result):
        ret = int(result[0])

        if ret != 0:
            return ret

        if self.term.next_binary(True) == None:
            print("No more files, exiting")
            return ret

        if (self.term.xfer.how == "xmodem"):
            self.sync("X")

        if not self.term.xfer.push(self.term.chip.spl_address):
            print("Upload failed")
            return 1
        return True

class flasher(basic_uploader):
    formats = {
        "flash_upload"      : "boot: Press '{}' and send me the image"
    }

    def action(self, trigger, result):
        self.term.xfer.selectTransport("xmodem-128")
        desc = "Writing image"
        self.sync("X")
        if not self.term.xfer.push(self.term.chip.spl_address):
            print("Upload failed")
            return 1
        return True


class base:
    name = "None"
    silent = False
    def __init__(self, opts):
        if opts.port[0].find("socket") != -1:
            self.silent = True

    def power(self, on):
        pass

    def resetWithCustomFlags(self, flags=[]):
        pass

    def resetToHost(self, flags = []):
        if not self.silent:
            print("Please, power-cycle board")

    def resetToNormal(self, flags = []):
        pass


from rumboot.classLoader import classLoader
from parse import *

class OpFactory(classLoader):
    objects = dict()
    def __init__(self, objectpath, terminal):
        super().__init__(objectpath)
        self.term = terminal
        for name,cl in self.classes.items():
            self.objects[name] = cl(terminal)

    def on_start(self):
        for name,obj in self.objects.items():
            obj.on_start()

    def handle_line(self, line, skipecho):
        for name,obj in self.objects.items():
            ret = obj.handle_line(line)
            if type(ret) is int:
                if not obj.hidden:
                    self.term.log(skipecho, line, end='\n')
                return ret
            if ret == True:
                if not obj.hidden:
                    self.term.log(skipecho, line, end='\n')
                return ret
        self.term.log(skipecho, line, end='\n')
        return False



from rumboot.classLoader import classLoader

class ResetSeqFactory(classLoader):
    def __getitem__(self, key):
        if key in self.classes:
            return self.classes[key]
        return self.classes["base"]

    def args(self, parser):
        pass




from rumboot.edclManager import edclmanager
from xmodem import XMODEM

class xferBase():
    connected = False
    chip = None
    def __init__(self, terminal):
        self.term = terminal
        pass

    def reconnect(self):
        pass
    
    def connect(self, chip):
        self.connected = True
        self.chip = chip
        return True

    def stream_size(self, stream):
        stream.seek(0,2)
        len = stream.tell()
        stream.seek(0)
        return len

    def write32(self, address, data):
        raise Exception("write32 not implemented")

    def read32(self, address):
        raise Exception("read32 not implemented")

    def _send(self, stream, destaddr, desc="Sending stream"):
        pass

    def send(self, stream, destaddr, desc="Sending stream"):
        if type(stream) is str:
            stream = open(stream, 'rb')
        if self.connected:
            return self._send(stream, destaddr, desc)
        else:
            raise Exception("Transport not connected")

    def _recv(self, stream, srcaddr, total, desc="Sending stream"):
        print("FATAL: NOT IMPLEMENTED")        
        pass

    def recv(self, stream, srcaddr, total, desc="Receiving stream"):
        if type(stream) is str:
            stream = open(stream, 'wb')
        if self.connected:
            return self._recv(stream, srcaddr, total, desc=desc)
        else:
            raise Exception("Failed to enable transport")

class xferXmodem(xferBase):
    increment = 128
    mode = "xmodem"

    def __init__(self, terminal):
        super().__init__(terminal)

    def connect(self, chip):
        ser = self.term.ser
        def getc(size, timeout=10):
            ret = ser.read(size)
            return ret or None
        def putc(data, timeout=10):
            return ser.write(data)  # note that this ignores the timeout
        self.modem = XMODEM(getc, putc, mode=self.mode)
        return super().connect(chip)

    def _send(self, stream, destaddr, desc="Sending stream"):
        total = self.stream_size(stream)
        terminal = self.term
        increment = self.increment
        terminal.progress_start(desc, total)
        self.last_ok = 0
        def callback(total_packets, success_count, error_count):
            if self.last_ok != success_count and success_count != 0:
                terminal.progress_update(total_packets * increment, success_count * increment, increment)
            self.last_ok = success_count
        ret = self.modem.send(stream, retry=128, callback=callback)
        terminal.progress_end()
        return ret
 
    def _recv(self, stream, srcaddr, total, desc='Receiving stream'):
        terminal = self.term
        increment = self.increment
        #terminal.progress_start(desc, total)
        self.last_ok = 0
        def callback(total_packets, success_count, error_count):
            if self.last_ok != success_count and success_count != 0:
                terminal.progress_update(total_packets * increment, success_count * increment, increment)
            self.last_ok = success_count
        #FixMe: modem.recv doesn't support callback mechanism
        ret = self.modem.recv(stream, crc_mode=0, retry=128)
        #terminal.progress_end()
        return ret

class xferXmodem1k(xferXmodem):
    mode = "xmodem1k"
    increment = 1024

class xferEdcl(xferBase):
    edcl = None
    force_static_arp = False
    def __init__(self, terminal):
        super().__init__(terminal)

    def connect(self, chip):
        if not self.edcl:
            self.edcl = edclmanager()
            self.edcl.force_static_arp = self.force_static_arp
            if not self.edcl.connect(chip):
                print("ERROR: Failed to establish edcl connection")
                return False
        return super().connect(chip)
        
    def _recv(self, stream, srcaddr, total, desc="Sending stream"):
        def prg(total_bytes, position, increment):
            self.term.progress_update(total_bytes, position, increment) 
        return self.edcl.recv_to_file(srcaddr, total, stream, callback=prg)

    def reconnect(self):
        if self.edcl != None:
            self.edcl.reconnect()

    def write32(self, address, data):
        return self.edcl.write32(address, data)

    def read32(self, address):
        return self.edcl.read32(address)

    def _send(self, stream, destaddr, desc='Sending stream'):
        terminal = self.term
        total = self.stream_size(stream)
        terminal.progress_start(desc, total)
        def prg(total_bytes, position, increment):
            terminal.progress_update(total_bytes, position, increment) 
        self.edcl.smartupload(destaddr, stream, callback=prg)
        terminal.progress_end()
        return True


class xferManager(xferBase):
    xfers = {}
    xfer = None
    how = "xmodem"

    def __init__(self, terminal, how="xmodem"):
        super().__init__(terminal)
        self.xfers["xmodem"] = xferXmodem1k(terminal)
        self.xfers["xmodem-128"] = xferXmodem(terminal)
        self.xfers["edcl"] = xferEdcl(terminal)
        self.how = how
        self.xfer = self.xfers[how]

    def setChip(self, chip):
        self.chip = chip

    def write32(self, address, data):
        return self.xfer.write32(address, data)

    def read32(self, address):
        return self.xfer.read32(address)
    
    def selectTransport(self, how):
        self.how = how
        if not how in self.xfers:
            raise Exception("Unknown xfer method: %s" % how)
        self.xfer = self.xfers[how]

    def push(self, address):
        binary = self.term.next_binary()
        return self.term.xfer.send(binary, address, "Sending binary")

    def send(self, stream, destaddr, desc='Sending stream'):
        if self.xfer.connect(self.chip):
            return self.xfer.send(stream, destaddr, desc=desc)
        else:
            raise Exception("Failed to connect transport")

    def recv(self, stream, srcaddr, total, desc="Receiving data"):
        if self.xfer.connect(self.chip):
            return self.xfer.recv(stream, srcaddr, total, desc=desc)
        else:
            raise Exception("Failed to connect transport")

    def reconnect(self):
        self.xfer.reconnect()
        
#    def _recv(self, stream, desc='Receiving stream'):
#        return super()._recv(stream, desc=desc)


import os
import time
import usb.core
import usb.util
from parse import parse
from rumboot.resetseq.resetSeqBase import base

class pl2303(base):
    name = "pl2303"
    invert = None
    pl2303_usb_dev = None


    class gpio:
        def __init__(self, dir_reg, dir_mask, out_reg, out_mask):
            self.dir_reg = dir_reg
            self.dir_mask = dir_mask
            self.out_reg = out_reg
            self.out_mask = out_mask


    # GPIO0: power
    GPIO_POWER = gpio(0x01, 0x10, 0x01, 0x40)

    # GPIO1: reset
    GPIO_RESET = gpio(0x01, 0x20, 0x01, 0x80)


    class usb_find_class(object):
        def __init__(self, usb_path):
            self.usb_path = usb_path

        def device_usb_path(self, dev):
            result = str(dev.bus) + "-" + str(dev.port_numbers[0])
            for i in range(1, len(dev.port_numbers)):
                result += "." + str(dev.port_numbers[i])
            return result

        def __call__(self, dev):
            return self.device_usb_path(dev) == self.usb_path


    def read_pl2303_reg(self, reg_num):
        buffer = self.pl2303_usb_dev.ctrl_transfer(
            usb.util.build_request_type(usb.util.CTRL_IN, usb.util.CTRL_TYPE_VENDOR, usb.util.CTRL_RECIPIENT_DEVICE),
            0x01,
            reg_num | 0x80,
            0,
            1,
            1000)
        return buffer[0]


    def write_pl2303_reg(self, reg_num, value):
        self.pl2303_usb_dev.ctrl_transfer(
            usb.util.build_request_type(usb.util.CTRL_OUT, usb.util.CTRL_TYPE_VENDOR, usb.util.CTRL_RECIPIENT_DEVICE),
            0x01,
            reg_num,
            value,
            None,
            1000)


    def set_gpio_out(self, gpio, v):
        if self.invert:
            v = not v

        dir_reg_value = self.read_pl2303_reg(gpio.dir_reg)
        dir_reg_value |= gpio.dir_mask
        self.write_pl2303_reg(gpio.dir_reg, dir_reg_value)

        out_reg_value = self.read_pl2303_reg(gpio.out_reg)
        if v:
            out_reg_value |= gpio.out_mask
        else:
            out_reg_value &= ~gpio.out_mask
        self.write_pl2303_reg(gpio.out_reg, out_reg_value)


    def __init__(self, opts):
        self.invert = opts.pl2303_invert

        if opts.pl2303_port: 
            # try to use a specianl port for PL2303
            tty = opts.pl2303_port[0]
        else:
            # if there is not the special port the main port will be used
            tty = opts.port[0]

        # canonze the port name
        tty = os.path.realpath(tty)
        parsed_tty_number = parse("/dev/ttyUSB{}", tty, case_sensitive=True)
        if not parsed_tty_number:
            raise Exception("cannot parse pl2303 port string: " + tty)
        tty = "ttyUSB" + str(parsed_tty_number[0])

        # get USB path for tty (like "1-1.1.3")
        sys_path = os.path.realpath("/sys/bus/usb-serial/devices/" + tty)
        usb_path = sys_path.split("/")[-2].split(":")[0]

        # find the device
        self.pl2303_usb_dev = usb.core.find(custom_match=self.usb_find_class(usb_path), idVendor=0x67b, idProduct=0x2303)
        if not self.pl2303_usb_dev:
            raise Exception("cannot find pl2303 device by usb path: " + usb_path)

        print("pl2303: %s detected at USB path %s" % (tty, usb_path))


    def resetWithCustomFlags(self, flags=[]):
        print("Please, power-cycle board")


    def power(self, on):
        if on:
            self.set_gpio_out(self.GPIO_POWER, 0)
        else:
            self.set_gpio_out(self.GPIO_POWER, 1)


    def resetToHost(self, flags = []):
        self.set_gpio_out(self.GPIO_POWER, 1)
        self.set_gpio_out(self.GPIO_RESET, 1)
        time.sleep(1)
        self.set_gpio_out(self.GPIO_RESET, 0)
        self.set_gpio_out(self.GPIO_POWER, 0)


    def resetToNormal(self, flags = []):
        self.set_gpio_out(self.GPIO_POWER, 1)
        self.set_gpio_out(self.GPIO_RESET, 1)
        time.sleep(1)
        self.set_gpio_out(self.GPIO_RESET, 0)
        self.set_gpio_out(self.GPIO_POWER, 0)


    def add_argparse_options(parser):
        parser.add_argument("-P", "--pl2303-port",
                            help="PL2303 physical port",
                            nargs=1, metavar=('value'),
                            required=False)
        parser.add_argument("--pl2303-invert",
                            help="Invert all pl2303 gpio signals",
                            action='store_true',
                            default = False,
                            required=False)



from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.LayoutFactory import LayoutFactory

import argparse
import rumboot_packimage
import rumboot

class RumbootPackimage:
    """RumbootPackimage image combiner"""
    def __init__(self, opts):
        pass

def cli():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-combine {} - RumBoot Image Merger Tool\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    parser.add_argument("-i", "--input",
                        help="Input image file (may be specified several times)",
                        type=argparse.FileType("r+b"),
                        action="append",
                        required=True)
    parser.add_argument("-o", "--output",
                        help="Output image",
                        required=True)
    parser.add_argument("-a", "--align",
                        default=1,
                        help='''Set alignment pattern of images in bytes or via keyword (SD, physmap, ini)
                        ''')


    opts = parser.parse_args()

    variant = LayoutFactory("rumboot.layouts")[opts.align](opts.output, opts.align)
    print("Using alignment scheme: " + variant.describe())

    for f in opts.input:
        variant.append(f)

    variant.close()
    print("Wrote " + opts.output + " successfully!")

    return 0



import sys
import time
from rumboot.resetseq.resetSeqBase import base

#   MT125.05. Shift register on CBUS pins
class mt12505(base):
    name = "MT125.05 (FT232RL)"
    flags = [ ]

    def reg_write_bit(self, bit):
        if bit>0:
            v = 0
        else:
            v = 1
        self.sp.cbus_write(v)    
        self.sp.cbus_write(v | (1<<3))
        self.sp.cbus_write(v)
        self.sp.cbus_write(0)

    def write_reg(self, values):
        self.sp.cbus_write(0)
        i=0
        for v in values:
            i=i+1
            self.reg_write_bit(v)
        self.sp.cbus_write(0)
        self.sp.cbus_write(1<<2)        
        self.sp.cbus_write(0)

#    def __init__(self, serial, flags = [0, 0, 0, 0, 0, 0, 0, 0], reset_pin = 5, host_pin = 3):
    def __init__(self, opts):
        import ft232
        flags = [0, 0, 0, 0, 0, 0, 0, 0]
        reset_pin = 5 
        host_pin = 3
        serial = opts.ft232_serial[0]        
        self.sp = ft232.Ft232(serial, baudrate=115200)
        self.sp.cbus_setup(mask=0xf, init=0xf)
        self.flags = flags
        self.reset_pin = reset_pin
        self.host_pin = host_pin

    def resetWithCustomFlags(self,flags = [0, 0, 0, 0, 0, 0, 0, 0]):
            tmp = flags
            tmp[self.reset_pin] = 0
            self.write_reg(tmp)
            time.sleep(1)
            tmp[self.reset_pin] = 1
            self.write_reg(tmp)
        
    def resetToHost(self, flags = [0, 0, 0, 0, 0, 0, 0, 0]):
            tmp = self.flags
            tmp[self.host_pin] = 1            
            tmp[self.reset_pin] = 0
            self.write_reg(tmp)
            time.sleep(1)
            tmp[self.reset_pin] = 1
            self.write_reg(tmp)
        
    def resetToNormal(self, flags = [0, 0, 0, 0, 0, 0, 0, 0]):
            tmp = self.flags
            tmp[self.host_pin] = 0            
            tmp[self.reset_pin] = 0
            self.write_reg(tmp)
            time.sleep(1)
            tmp[self.reset_pin] = 1
            self.write_reg(tmp)

    def add_argparse_options(parser):
        parser.add_argument("-S", "--ft232-serial",
                            help="FT232 serial number for MT125.05",
                            nargs=1, metavar=('value'),
                            default=["A92XPFQL"],
                            required=False)




import os
import heatshrink2
from rumboot.images.imageFormatBase import ImageFormatBase

class ImageFormatV3(ImageFormatBase):
    """
    This class works with version 3.0 images. 
    3.0 are backwars compatible with V2

    enum rumboot_header_flags {
        RUMBOOT_FLAG_COMPRESS = (1 << 0), /** Image data is compressed with heatshrink */
        RUMBOOT_FLAG_DATA     = (1 << 1), /** Image file contains only data to be loaded, it shouldn't be executed */
        RUMBOOT_FLAG_RESERVED = (1 << 2), /** Reserved for further usage */ 
        RUMBOOT_FLAG_SMP      = (1 << 3), /** SMP Image. If supported by target CPU cluster, all cores will start execution */
        RUMBOOT_FLAG_DECAPS   = (1 << 4), /** Remove header before executing and move data to the beginning */
        RUMBOOT_FLAG_RELOCATE = (1 << 5), /** Relocate image before execution to the address in relocation field */
        RUMBOOT_FLAG_SYNC     = (1 << 6), /** Wait for the image to finish before exiting */
        RUMBOOT_FLAG_KILL     = (1 << 7), /** Issue a reset to the target CPU before operating */
    };

    struct __attribute__((packed)) rumboot_bootheader {
        uint32_t magic;
        uint8_t  version;
        uint8_t  flags;
        uint8_t  chip_id;
        uint8_t  chip_rev;
        uint32_t data_crc32;
        uint32_t datalen;

        union
        {
          uint64_t  entry_point;
          uint32_t  entry_point32[2];
        };

        uint64_t  relocation;
        uint32_t  target_cpu_cluster;
        uint32_t  encryption_slot;
        uint32_t  certificate_slot;
        uint32_t  priority;
        uint32_t  reserved[2];
        uint32_t  header_crc32;
        const struct rumboot_bootsource *device;
        char     data[];
    };

    """
    name = "RumBootV3"
    MAGIC = 0xb01dface
    VERSION = 3

    flags = ["COMPRESS", "DATA", "SIGN", "SMP", "DECAPS", "RELOC", "SYNC", "KILL"]
    _flags = {}
    format = [
        [4, "magic",             "0x%x", "Magic"],
        [1, "version",           "0x%x", "Header Version"],
        [1, "flags",             "0x%x", "Flags"],
        [1, "chip_id",           "0x%x", "Chip ID"],
        [1, "chip_rev",          "0x%x", "Chip Revision"],
        [4, "data_crc32",        "0x%x", "Data CRC32"],
        [4, "data_length",       "%d",   "Data Length"],
        [8, "entry_point",       "0x%x", "Entry Point"],
        [8, "relocation",        "0x%x", "Relocation Address"],
        [1, "target_cpu",        "0x%x", "Target CPU Cluster"],
        [1, "encryption_slot",   "0x%x", "Encryption Slot"],
        [1, "certificate_slot",  "0x%x", "Cetificate Slot"],
        [1, "priority",          "0x%x", "Priority"],
        [8, "bootarg0",          "0x%x", "Boot Arg 0"],
        [8, "bootarg1",          "0x%x", "Boot Arg 1"],
        [4, "reserved",          "0x%x", ""],
        [4, "header_crc32",      "0x%x", "Header CRC32"],
        [4, "bootsource",        "0x%x", ""],
    ]

    header = {}

    def read_header(self):
        super().read_header()
        self.read_flags()

    def read_flags(self):
        for k,f in enumerate(self.flags):
            if (self.header["flags"] & (1<<k)):
                self._flags[f] = True
            else:
                self._flags[f] = False

    def write_flags(self):
        the_flags = 0
        for k,f in enumerate(self.flags):
            if self._flags[f]:
                the_flags |= 1<<k
        self.header["flags"]=the_flags
        self.read_flags()

    def serialize_flags(self):
        ret = ""
        total = 0
        for k,f in enumerate(self.flags):
               if self._flags[f]:
                if total>0:
                   ret += " "
                ret += f
                total += 1
        return ret

    def flag(self, flag, value=None):
        if not isinstance(value, bool):
            return self._flags[flag]
        else:
            self._flags[flag] = value
            self.write_flags()

    def relocate(self, address):
        self.flag("RELOC", True)
        self.header["relocation"] = int(address, 16)
        self.write_header()

    def dump_header(self, raw=False, format=False):
        self.read_flags()
        self.hide_field("flags")
        self.hide_field("encryption_slot")
        self.hide_field("certificate_slot")
        if not self.flag("RELOC"):
            self.hide_field("relocation")
        if self.header["priority"] == 0:
            self.hide_field("priority")
        self.hide_field("reserved")
        ret = super().dump_header(raw, format)
        self.dump_field("flags", False, self.serialize_flags(), raw)
        return ret;

    def check(self):
        if (super().check()):
            if self.header["version"] == self.VERSION:
                self.read_flags()
                return True
        return False

    def __init__(self, inFile):
        super().__init__(inFile)

    def wrap(self):
        super().wrap()
        self.header["version"] = 3
        self.write_header()
        return True

    def get_chip_id(self):
        return self.header['chip_id']

    def get_chip_rev(self):
        return self.header['chip_rev']

    def fix_checksums(self, calc_data = True):
       self.write_flags()
       super().fix_checksums(calc_data)

    def compress(self):
        if self.flag("COMPRESS"):
            print("WARN: Image already compressed!")
            return
        uncompressed_length = self.file_size - self.get_header_length()
        self.fd.seek(self.get_header_length(), os.SEEK_SET)
        data = self.fd.read(uncompressed_length)
        data = heatshrink2.compress(data, window_sz2=11, lookahead_sz2=4)
        self.fd.seek(self.get_header_length(), os.SEEK_SET)
        self.fd.truncate(self.get_header_length())
        self.fd.write(data)
        self.flag("COMPRESS", True)
        self.read_file_size()

    def decompress(self):
        if not self.flag("COMPRESS"):
            print("WARN: Image not compressed, can't decompress!")
            return
        compressed_length = self.file_size - self.get_header_length()
        self.fd.seek(self.get_header_length(), os.SEEK_SET)
        data = self.fd.read(compressed_length)
        data = heatshrink2.decompress(data,window_sz2=11, lookahead_sz2=4)
        self.fd.seek(self.get_header_length(), os.SEEK_SET)
        self.fd.truncate(self.get_header_length())
        self.fd.write(data)
        self.flag("COMPRESS", False)
        self.read_file_size()



import serial
import sys
from xmodem import XMODEM
import os
from parse import parse
import time
import io
from tqdm import tqdm
from rumboot.OpFactory import OpFactory
import threading
import serial
import serial.rfc2217
import socket
import select
import signal

class redirector(threading.Thread):
    alive = False
    fatal = False
    callback = None

    def configure(self, serial, socket):
        self.serial = serial
        self.socket = socket

    def set_callback(self, cb):
        self.callback = cb

    def cleanup(self, fatal):
        self.socket.close()
        self.fatal = fatal
        self.alive = False
        if not fatal:
            print("Client disconnected")
        else:
            print("Something bad happened. Stopping daemon")
        if self.callback != None:
            print("calling", fatal)
            self.callback(fatal)

    def run(self):
        self.alive = True
        self.fatal = False
        self.socket.setblocking(0)

        fromserial = bytearray(b"")
        fromsocket = bytearray(b"")
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        self.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 1)
        self.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 1)
        self.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 2)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        try:
            while True:
                ready_to_read, ready_to_write, in_error = select.select(
                    [self.socket, self.serial.fileno()],
                    [self.socket, self.serial.fileno()],
                    [self.socket, self.serial.fileno()],
                    1)


                for sock in ready_to_read:       
                    if sock == self.serial.fileno():
                        fromserial = fromserial + bytearray(self.serial.read(self.serial.inWaiting()))
                        if self.socket in ready_to_write:
                            sent = self.socket.send(fromserial)
                            del fromserial[0:sent]

                    if sock == self.socket:
                        tmp = self.socket.recv(1024)
                        if len(tmp) == 0:
                            self.cleanup(False)
                            return
                        fromsocket = fromsocket + bytearray(tmp)
                        if self.serial.fileno() in ready_to_write:
                            ret = self.serial.write(fromsocket)
                            del fromsocket[0:ret]

                for sock in in_error:
                    if sock == self.serial.fileno():
                        print("Something bad with serial port")
                        self.cleanup(True)
                        return
                    if sock == self.socket:
                        print("Disconnect?")
                        self.cleanup(False)               
                        return
        except BrokenPipeError:
            self.cleanup(False)
            return
        except ConnectionResetError:
            self.cleanup(False)
            return
        except OSError:
            self.socket.close()
            self.cleanup(True)
            raise
        except Exception:
            self.cleanup(False)
            raise

class server:
    client_queue = [ ]
    worker = None
    #Graveyard of zombies
    graveyard = []
    pid = os.getpid()
    binaries = []

    def __init__(self, terminal, tcplisten):
        self.serial = terminal.serial()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        addr, port = tcplisten.split(":")
        self.sock.bind((addr, int(port)))
        self.term = terminal
        terminal.chip.hacks["skipsync"] = True

    def set_reset_seq(self, rst):
        self.rst = rst

    def preload_binaries(self, files):
        self.binaries = files

    def serve_once(self):
        def the_callback(fatal = False):
            if not fatal:
                self.serve_once()
            else:
                print("Interrupting main thread")
                os.kill(self.pid, signal.SIGINT)

        if self.worker != None:
            if self.worker.alive:
                return #We're busy here
            else:
                #Put our dead worker to the graveyard
                self.graveyard = self.graveyard + [ self.worker ]

        try:
            client = self.client_queue.pop(0)
            # Reset if using a nested damon
            self.term.reopen()
            self.serial = self.term.serial()
            # Reset if using a local resetter
            self.rst.resetToHost()
            if self.binaries:
                text = b"U\nrumboot-daemon: Preloading your board board...\n\n\n"
                client["connection"].sendall(text)

                print(self.binaries)
                self.term.add_binaries(self.binaries)
                self.term.loop(break_after_uploads=True)

            self.worker = redirector()
            self.worker.configure(self.serial, client["connection"])
            self.worker.set_callback(the_callback)

            # We can't do it, if we're working remotely
            # Or somebody resets us manually
#            if type(self.rst).__name__ != "base":
#                self.serial.reset_input_buffer()
#                self.serial.reset_output_buffer()
            
            self.worker.start()
            print("Now serving client: ", client["dns"])
        except(IndexError):
            self.worker = None
            self.rst.power(0) # Power off board

    def queue_client(self, connection, client_address):
        try:
            dns = socket.gethostbyaddr(client_address[0])
        except:
            dns = "<unknown>"
        print("Incoming connection: ", dns)
        client = { }
        client["connection"] = connection
        client["address"] = client_address
        client["dns"] = dns
        self.client_queue.append(client)
        pos = len(self.client_queue)
        if self.worker != None:
            pos = pos + 1

        text = b"U\nrumboot-daemon: You are client number %d in queue, please stand by\n\n\n" % pos
        connection.sendall(text)

        if self.worker == None:
            self.serve_once()


    def kill_zombies(self):
        #Kill all zombies.
        for z in self.graveyard:
            z.join()
        self.graveyard = [ ]                    

    def cleanup(self):
        if self.worker:
            self.worker.join()
        self.worker.socket.close()
        self.sock.close()

    def loop(self):
        self.rst.power(0) # Power off board
        try:
            self.sock.listen()
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 1)
            self.sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 1)
            self.sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 2)

            while True:
                print('waiting for a connection')
                connection, client_address = self.sock.accept()
                self.queue_client(connection, client_address)
                self.kill_zombies()
        except:
            self.cleanup()
        finally:
            self.cleanup()



from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal

import os
import argparse
import rumboot_xflash
import rumboot_packimage
from parse import *
import rumboot
import threading
import socket
from rumboot.server import redirector


class ipflashrom(threading.Thread):
    def configure(self, flashrom, port, args):
        self.flashrom = flashrom
        self.port = port
        self.args = args

    def run(self):
        cmd = self.flashrom + " -p serprog:ip=127.0.0.1:" + str(self.port) + " " + self.args
        #cmd = "nc 127.0.0.1 "+ str(self.port) 
        print("Invoking flashrom: " + cmd)
        return os.system(cmd)


def cli():
    resets  = ResetSeqFactory("rumboot.resetseq")
    formats = ImageFormatDb("rumboot.images")
    chips   = ChipDb("rumboot.chips")

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-flashrom {} - flashrom wrapper tool\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    helper = arghelper();                                
    helper.add_terminal_opts(parser)
    parser.add_argument("-v", "--verbose", 
                        action='store_true',
                        default=False,
                        help="Print serial debug messages during preload phase"),
    parser.add_argument("-m", "--memory",
                        help="SPI bus to use. Names match '-m help' of rumboot-xflash",
                        nargs=1, metavar=('memory'),
                        default="help",
                        required=True)                        
    parser.add_argument("-z", "--spl-path",
                        help="Path for SPL writers (Debug only)",
                        type=str,
                        required=False)
    parser.add_argument("-f", "--flashrom-path",
                        help="Path to flashrom binary",
                        type=str,
                        required=False)
    parser.add_argument("-c", "--chip_id",
                    help="Chip Id (numeric or name)",
                    required=True)        
    parser.add_argument('remaining', nargs=argparse.REMAINDER, default=[], help="Flashrom arguments")
#    parser.add_argument('-F', '--flashrom-args', nargs='*', default=[], )

    helper.add_resetseq_options(parser, resets)

    opts = parser.parse_args()
    if opts.remaining and opts.remaining[0] == "--":
        opts.remaining = opts.remaining[1:]
    flashrom_args = " ".join(opts.remaining)

    c = chips[opts.chip_id]
    if (c == None):
        return 1;
 
    helper.detect_terminal_options(opts, c)
    print("Detected chip:    %s (%s)" % (c.name, c.part))

    if c.warning != None:
        print("    --- WARNING ---")
        print(c.warning)        
        print("    --- WARNING ---")
        print("")

    spl_path = os.path.dirname(__file__) + "/tools/"

    if opts.baud == None:
        opts.baud = [ c.baudrate ]

    reset = resets[opts.reset[0]](opts)
    term = terminal(opts.port[0], opts.baud[0])
    term.verbose = opts.verbose
    term.set_chip(c)
    if opts.log:
        term.logstream = opts.log
    
    mem = opts.memory[0]
    if not mem:
        return 1

    if mem == "help":
        for mem,spl in c.flashrom.items():
            print("Memory %16s: %s" % (mem, spl))
        return 1

    try:
        spl = c.flashrom[mem]
    except:
        print("ERROR: Target chip (%s) doesn't have bus '%s'. Run with -m help for a list" % (c.name, mem))
        return 1

    if opts.spl_path != None:
        spl_path = opts.spl_path
        print("SPL               %s" % (spl_path + c.flashrom[mem]))

    spl = spl_path + c.flashrom[mem]

    print("Reset method:     %s" % (reset.name))
    print("Baudrate:         %d bps" % int(opts.baud[0]))
    print("Port:             %s" % opts.port[0])

    flashrom = None

    flashrom_paths = [
        "/usr/local/sbin/flashrom",
        "/sbin/flashrom", 
        "/usr/sbin/flashrom",
        "C:\\flashrom\\flashrom.exe"
        ]

    if opts.flashrom_path and os.path.exists(opts.flashrom_path):
        flashrom = opts.flashrom_path
    elif opts.flashrom_path:
        print("[W] Specified flashrom path (%s) is invalid, trying to guess" % opts.flashrom_path)

    for f in flashrom_paths:
        if os.path.exists(f):
            flashrom = f

    if flashrom == None:
        print("FATAL: Failed to find working flashrom. Please download and install from https://flashrom.org/")
        return 1

    print("FlashRom:         %s" % flashrom)
    print("FlashRom args:    %s" % flashrom_args)

    reset.resetToHost()
    term.add_binaries(spl)

    term.loop(break_after_uploads=True)
    while True:
        s = term.ser.readline()
        if s.decode("utf-8", errors="replace").find("SERPROG READY!") != -1:
            break

    print("Serprog stub ready!")

    if opts.port[0].find("socket://") == -1:
        #Use flashrom directly

        ret = os.system(flashrom + " -p serprog:dev="+opts.port[0] + ":" + opts.baud[0])
        reset.resetToNormal
        return ret


    # If we're here, we'll need to setup socket redirector
    port=20000

    while True:
        try:
            print("Trying port %d" % port)
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.bind(("127.0.0.1", port))
            server.listen()            
            break
        except:
            port = port + 1
            if port > 30000:
                print("Weird shit going on, breaking")
                break

    #Fire up flashrom in background
    flr = ipflashrom()
    flr.configure(flashrom, port, flashrom_args)
    flr.start()

    #Accept connection
    connection, client_address = server.accept()
    connection.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    rd = redirector()
    rd.configure(term.ser, connection)
    rd.start()

    return flr.join()
    






from rumboot.ImageFormatDb import ImageFormatDb
import argparse
import rumboot_packimage
import rumboot

class RumbootPackimage:
    """RumbootPackimage tool frontend"""
    def __init__(self, opts):
        pass

def cli():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-packimage {} - Universal RumBoot Image Manipulation Tool\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    parser.add_argument("-f", "--file",
                        help="image file",
                        type=argparse.FileType("r+b"),
                        required=True)
    parser.add_argument("-i", "--info",
                        help="Show information about the image",
                        action="store_true")
    parser.add_argument("-c", "--checksum",
                        help='''This option will modify the file! Calculates valid checksums to the header.
                        The length is set to cover the full length of file ONLY if it's zero. 
                        ''',
                        action="store_true")
    parser.add_argument("-C", "--checksum_fix_length",
                        help="This option will modify the file! The same as --checksum/-c, but always overrides length to covert the full length of file",
                        action="store_true")
    parser.add_argument("-r", "--raw",
                        help="Display raw header field names",
                        action="store_true")
    parser.add_argument('-R', '--relocate',
                        nargs=1, metavar=('relocation'),
                        help='''Tell bootrom to relocate the image at the specified address before executing it. Only RumBootV3 and above
                        ''')
    parser.add_argument('-Z', '--compress',
                        action="store_true",
                        help='''Compress image data with heatshrink algorithm (V3 or above only)
                        ''')
    parser.add_argument('-U', '--decompress',
                        action="store_true",
                        help='''Decompress image data with heatshrink algorithm (V3 or above only)
                        ''')
    parser.add_argument('-z', '--add_zeroes',
                        nargs=1, metavar=('value'),
                        help='''This option will add N bytes of zeroes to the end of the file (after checksummed area).
                        This is required to avoid propagating 'X' during the next image check during simulation.
                        Normally, you do not need this option
                        ''')
    parser.add_argument('-a', '--align',
                        nargs=1, metavar=('value'),
                        help='''Pad resulting image size to specified alignment. Remember to add -C to have correct checksums!
                        ''')
    parser.add_argument('-F', '--flag',
                        action="append",
                        nargs=2, metavar=('value'),
                        help='''Set image flag to a desired value. Only RumBootV3 or above
                        ''')
    parser.add_argument('--set-data',
                        action="append",
                        nargs=2, metavar=('offset', 'value'),
                        help='''Sets data at byte 'offset' to value 'offset'
                        ''')
    parser.add_argument('-g', '--get',
                        nargs=1, metavar=('key'),
                        help='''Get a single field from header. Nothing else will be printed.
                        NOTE: The value will be formatted as hex
                        ''')
    parser.add_argument('-s', '--set',
                        nargs=2, metavar=('key', 'value'),
                        help='''This option will modify the file! Set a header key to specified value.
                                Use -r flag on an existing image to find out what keys exist.
                                Use -c to update the checksums
                        ''')
    parser.add_argument('-e', '--reverse-endianness',
                        action="store_true",
                        help='''Use this option to reverse endianness of all headers. This will not touch data.
                                For testing only
                        ''')
    parser.add_argument('-w', '--wrap',
                        nargs=1,
                        help='''Use this option to wrap arbitrary data to V1/V2/V3 images.
                        ''')

    opts = parser.parse_args()
    formats = ImageFormatDb("rumboot.images");
    t = formats.guess(opts.file)

    if opts.wrap:
        if t != False:
            print(f"File {opts.file.name} already wrapped into rumboot format ({t})")
            return 1            
        t = formats.wrap(opts.file, opts.wrap[0])
        print(f"Wrapped file {opts.file.name} with {t.name} header")

    calc_data = True
    if (t == False):
        if not opts.wrap:
            print("ERROR: Not a valid image, see README.md")
        else:
            print("ERROR: Failed to wrap data into image for some reason")
        return 1

    if opts.set != None:
        t.set(opts.set[0], opts.set[1])

        #FixMe: Hack
        if (opts.set[0] == "data_crc32"):
            calc_data = False
        opts.info = True
        print("Setting " + opts.set[0] + " to " + opts.set[1])
        if not opts.checksum:
            print("WARNING: Add -c option to update checksums!")

    if opts.compress:
        if not hasattr(t, "compress"):
            print("ERROR: Image (de)compression is not supported by this image format")
            return 1
        t.compress()
        opts.checksum_fix_length = True
        opts.checksum = True
        opts.info = True

    if opts.decompress:
        if not hasattr(t, "decompress"):
            print("ERROR: Image (de)compression is not supported by this image format")
            return 1
        t.decompress()
        opts.checksum_fix_length = True
        opts.checksum = True
        opts.info = True

    if opts.relocate:
        if not hasattr(t, "relocate"):
            print("ERROR: Relocation is not supported by this image format")
            return 1
        t.relocate(opts.relocate[0])
        opts.info = True
        if not opts.checksum:
            print("WARNING: Add -c option to update checksums!")

    if opts.get:
        print("0x%x" % t.get(opts.get[0]))
        return 0

    if opts.reverse_endianness:
        t.swap_endian()

    if opts.set_data:
        for f in opts.set_data:
            t.write8(t.get_header_length() + int(f[0]), int(f[1]))

    if opts.flag:
        for f in opts.flag:
            if not hasattr(t, "flag"):
                print("ERROR: Image flags are not supported by this image format")
                return 1
            t.flag(f[0],bool(f[1]))

    if opts.align:
        t.align(opts.align[0])
        opts.checksum_fix_length = True

    if opts.add_zeroes:
        t.add_zeroes(opts.add_zeroes[0])

    if (opts.checksum_fix_length):
        t.fix_length()
        opts.info = True
    elif opts.checksum:
        t.fix_checksums(calc_data)
        print("Wrote valid checksums to image header")
        opts.info = True



    if opts.info:
        t.read_header()
        #Return non-zero on invalid data/header crc, except for error-injection cases
        if not t.dump_header(opts.raw) and not opts.set:
            return 1

    return 0



import os
import time
from rumboot.resetseq.resetSeqBase import base

class powerhub(base):
    name = "PowerHub"
    port = 0
    def __init__(self, opts):
        pass

    def power(self, on):
        if on:
            os.system("powerhubctl outlet_power_ctl 0");
        else:
            os.system("powerhubctl outlet_power_ctl 1");


    def resetWithCustomFlags(self, flags=[]):
        print("Please, power-cycle board")

    def resetToHost(self, flags = []):
        os.system("powerhubctl outlet_power_ctl 0");
        time.sleep(1)
        os.system("powerhubctl outlet_power_ctl 1");

    def resetToNormal(self, flags = []):
        os.system("powerhubctl outlet_power_ctl 0");
        time.sleep(1)
        os.system("powerhubctl outlet_power_ctl 1");


from rumboot.images.imageFormatBase import ImageFormatBase

class ImageFormatV1(ImageFormatBase):
    """
    This class works with 1888TX018 (V1) images
    struct bootheader
    {
        uint32_t magic;
        uint32_t length;
        uint32_t entry0;
        uint32_t entry1;
        uint32_t sum;
        uint32_t hdrsum;
        uint8_t imagedata[]; /* Image data. */
    } __attribute__((packed));

    """
    name = "RumBootV1"
    MAGIC = 0xdeadc0de
    format = [
        [4, "magic", "0x%x", "Magic"],
        [4, "data_length", "%d", "Data Length"],
        [4, "entry0",  "0x%x", "Entry Point[0]"],
        [4, "entry1",  "0x%x", "Entry Point[1]"],
        [4, "data_crc32", "0x%x", "Data CRC32"],
        [4, "header_crc32", "0x%x", "Header CRC32"],
    ]

    def __init__(self, inFile):
        super().__init__(inFile)

    def get_chip_id(self):
        return 1

    def get_chip_rev(self):
        return 1 

    def wrap(self):
        return super().wrap()



from rumboot.ops.base import base
from parse import parse
import sys
class stackframe(base):
    hidden = True
    dumps = {}
    formats = {
        "mcsrr0": "MCSRR0: {}",
        "csrr0":  "CSRR0: {}",
        "srr0":   "SRR0: {}",
        "frame" : "frame[{}] address {}",
        }

    def getdump(self, dmp):
        if dmp == None:
            return { }
        if dmp.name in self.dumps:
            return self.dumps[dmp.name]
        dmp.seek(0)
        dumpdb = {}
        func = "none"
        funcaddr = 0
        while True:
            try:
                line = dmp.readline()
                if not line:
                    break
            except:
                continue
            ret = parse("{} <{}>:", line)
            if ret:
                func = ret[1]
                funcaddr = int(ret[0],16)

            if func.find(".debug_") >= 0:
                continue

            line = line.strip().split("\t")
            try:
                add = int(parse("{}:", line[0])[0], 16)
                opcodes = line[2]
                del line[0:2]
                instructions = " ".join(line)
                dumpdb[add] = {}
                dumpdb[add]["assembly"]       =  instructions
                dumpdb[add]["opcodes"]        = opcodes
                dumpdb[add]["function"]       = func
                dumpdb[add]["function_addr"]  = funcaddr
            except:
                pass
            # Cache it!
        self.dumps[dmp.name] = dumpdb
        return dumpdb


    def lookup(self, address):
        dump = self.getdump(self.term.current_dump())
        if not address in dump:
            dump = self.getdump(self.term.dumps["rom"])
        if not address in dump:
            return None
        return dump[address]
        
    def dump_register(self, name, address):
        if address == 0:
            return False
        data = self.lookup(address)
        if data:
            self.term.log(False, "%s:  0x%x [%s(): %s]" % (name, address, data["function"], data["assembly"]))
            return True

    def action(self, trigger, result):
        if trigger == "frame":
            id = int(result[0])
            address = int(result[1], 16)
            data = self.lookup(address)
            if data != None:
                self.term.log(False, "%d. 0x%.08x %30s(): %s" % (id, address, data["function"], data["assembly"]))
                return True
            else:
                self.term.log(False, "%d. 0x%.08x %30s(): INVALID OR MISSING FRAME" % (id, address, "???"))
                return True
        if trigger == "mcsrr0":
            address = int(result[0].strip(), 16)
            return self.dump_register("MCSRR0", address)
        if trigger == "csrr0":
            address = int(result[0].strip(), 16)
            return self.dump_register("CSRR0", address)
        if trigger == "srr0":
            address = int(result[0].strip(), 16)
            return self.dump_register("srr0", address)





import serial
import time

class redd:
    name = "redd"
    silent = False
    magicsymbol = 'a'
    ser = None
    def __init__(self, opts):
        self.magicsymbol = opts.redd_relay_id
        self.ser = serial.Serial(opts.redd_port)         

    def power(self, on):
        if on:
            self.ser.write(self.magicsymbol.capitalize().encode())
        else:
            self.ser.write(self.magicsymbol.lower().encode())

    def resetWithCustomFlags(self, flags=[]):
        self.power(0)
        time.sleep(1)
        self.power(1)

    def resetToHost(self, flags = []):
        self.power(0)
        time.sleep(1)
        self.power(1)

    def resetToNormal(self, flags = []):
        self.power(0)
        time.sleep(1)
        self.power(1)

    def add_argparse_options(parser):
        parser.add_argument("--redd-port",
                            help="Redd serial port (e.g. /dev/ttyACM1)",
                            default="/dev/ttyACM1",
                            required=False)
        parser.add_argument("--redd-relay-id",
                            help="Redd Relay Id (e.g. A)",
                            default="B",
                            required=False)


import serial
import sys
import os
from parse import parse
import time
import io
from tqdm import tqdm
from rumboot.OpFactory import OpFactory
from rumboot.chips.base import chipBase
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.xfer import xferManager
import socket
import select

if sys.platform != "win32":
	import tty
	import termios

class terminal:
        verbose=True
        logstream=None
        plusargs={}
        initial_loop_done = False
        runlist = []
        formats = ImageFormatDb("rumboot.images")
        chip = chipBase()
        dumps = {}
        curbin = "bootrom"
        progress = tqdm(disable=True)
        replay = False
        replay_till_the_end = False
        ser = None
        cmdline = None
        xfer = None
        desc_widget = None
        prog_widget = None
        shell_prompt = None
            
        def __init__(self, port, speed):
            self.port = port
            self.speed = speed
            self.xfer = xferManager(self)
            self.reopen()

        def serial(self):
            return self.ser

        def reopen(self):
            if self.ser != None:
                self.ser.close()
                self.ser = None

            if self.port.find("file://") == 0:
                self.ser = open(self.port.replace("file://",""), "rb+")
                self.replay = True
            elif self.port.find("://") < 0:
                self.ser = serial.Serial(self.port, self.speed, timeout=5)
            else:
                self.ser = serial.serial_for_url(self.port, timeout=5)
            self.opf = OpFactory("rumboot.ops", self) 


        def set_chip(self, chip):
            self.chip = chip
            self.xfer.setChip(chip)

        def tqdm(self, *args, **kwargs):
            self.progress.close()
            self.progress = tqdm(*args, **kwargs)

        def add_binaries(self, path):
            if type(path) is list:
                for p in path:
                    self.runlist.append(p[0])
            else:
                self.runlist.append(path)

        def add_dumps(self, dumps):
            self.dumps.update(dumps)

        def next_binary(self, peek=False):
            if len(self.runlist) > 0:
                ret = self.runlist[0]
                if not peek:
                    self.runlist.pop(0)
                    self.curbin = ret
                return ret
            return None

        def current_binary(self):
            return self.curbin

        def current_dump(self):
            if self.curbin == "bootrom":
                return self.dumps["rom"]
            if self.curbin.name in self.dumps:
                return self.dumps[self.curbin.name]

        def log(self, skipecho, *args, **kwargs):
            # Sometimes we get a weird exception on certain windows systems/setups
            # Try to catch any exceptions and dispose of them here
            try:
                if not skipecho and self.verbose:
                    self.progress.write(*args, **kwargs)
                    sys.stdout.flush()
                if not self.logstream == None:
                    self.logstream.write(*args)
                    self.logstream.write("\n")
            except:
                pass


        def sync(self):
            if not self.replay:
                self.ser.reset_input_buffer()
            c1 = b'a'
            c2 = b'd'
            while True:
                c2 = self.ser.read(1)
                if c1 == b'U' and c2 == b'\r':
                    break
                if c1 == b'U' and c2 == b'\n':
                    break
                c1 = c2

        def hack(self, name):
            if name in self.chip.hacks and self.chip.hacks[name]:
                return True
            return False

        def set_progress_widgets(self, desc_widget = None, prog_widget = None):
            self.desc_widget = desc_widget
            self.prog_widget = prog_widget

        def progress_start(self, description, total):
            if self.prog_widget == None:
                self.tqdm(desc=description, total=total, unit_scale=True, unit_divisor=1024, unit='B', disable=False)
            else:
                pass

        def progress_update(self, total, position, increment):
            self.progress.update(increment)

        def progress_end(self):
            self.tqdm(disable=True)

        def wait(self, format, timeout=1000):
            while True:
                line = self.ser.read_until()
                line = line.decode(errors="replace").rstrip()
                self.log(False, line, end='\n')
                ret = parse(format, line)
                if ret != None:
                    return ret

        def shell_mode(self, prompt):
            self.shell_prompt = prompt

        def wait_prompt(self, initial=False):
            if initial:
                self.ser.write(b"\r\n")
            found = False
            data = b""
            while True:
                ret = self.ser.read_until(self.shell_prompt.encode())
                data = data + ret
                ret = ret.decode(errors="replace")
                self.log(False, ret, end='\n')
                if ret.find(self.shell_prompt) > -1:
                    found = True
                    if not initial:
                        break
                if found and len(ret) == 0:
                    break
            return data

        def shell_cmd(self, cmd, timeout=10, initial=False):
            if initial:
                self.wait_prompt(initial)
            cmd = cmd.encode() + b"\r"
            self.ser.write(cmd)
            ret = self.wait_prompt(False).decode(errors="replace")
            ret = ret[len(cmd):] 
            ret = ret[:-len(self.shell_prompt)-2]
            return ret                  

        def loop(self, use_stdin=False, break_after_uploads=False, timeout=-1):
            if not self.initial_loop_done:
                self.initial_loop_done = True
                self.opf.on_start()

            return_code = 0
            time_start = time.time()
            if use_stdin:
                old_settings = termios.tcgetattr(sys.stdin)
            try:
                if use_stdin:
                    tty.setcbreak(sys.stdin.fileno())

                if not self.hack("skipsync"):
                    self.sync()
                while True:
                    ret = None
                    if use_stdin : 
                        line = b''
                        sym = b''
                        while sym != b'\n':
                            if select.select([self.ser], [], [], 0) == ([self.ser], [], []):
                                sym = self.ser.read(1)
                                line = line + sym
                                sys.stdout.write(str(sym, 'utf-8', errors='replace'))
                                sys.stdout.flush()
                            if select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], []):
                                insym = sys.stdin.read(1)
                                self.ser.write(insym.encode())
                    else:
                        line = self.ser.readline()

                    if timeout >= 0 and time.time() - time_start > timeout:
                        return_code = 5
                        break

                    if line == b'':
                        continue
                    try: 
                        line = line.decode().rstrip()
                    except:
                        continue

                    ret = self.opf.handle_line(line, use_stdin)
                    if type(ret) is int:
                        if not self.replay_till_the_end:
                            return ret
                        else:
                            return_code = ret
                    
                    if break_after_uploads and len(self.runlist) == 0:
                        break

                    if self.replay:
                        if (self.ser.tell() == os.fstat(self.ser.fileno()).st_size):
                            print("== END OF LOG REACHED ==")
                            break
            finally:
                if use_stdin:
                    termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)
                    return return_code



import os
from time import sleep
import socket
from rumboot.resetseq.resetSeqBase import base

class apc(base):
    name = "apc"
    host = "192.168.10.2"
    port = 23
    user = "whiteblade"
    passwd = "ash2wert"
    outlet = 8

    def __init__(self, opts):
        self.passwd = opts.apc_pass
        self.host   = opts.apc_host
        self.user   = opts.apc_user
        self.outlet = opts.apc_outlet


    def expect(self, f, inw, outw):
        failures = 0
        while True:
            if failures > 30:
                raise Exception("Problem with APC")
            tmp = f.readline()
            if len(tmp) == 0:
                sleep(0.1)
                failures += 1
                pass
            if inw in tmp:
                f.write(outw + "\r")
                f.flush()
                return

    def _power(self, on):
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                 s.connect((self.host, self.port))
                 s.setblocking(0)
                 if on:
                     on = 1
                 else:
                     on = 2
                 f = s.makefile('rw', encoding="ascii", errors="replace", newline='\r')
                 self.expect(f, "User Name :", self.user)
                 self.expect(f, "Password  :", self.passwd)
                 self.expect(f, ">", "1")
                 self.expect(f, ">", "2")
                 self.expect(f, ">", "1")
                 self.expect(f, ">", str(self.outlet))
                 self.expect(f, ">", str(on))
                 self.expect(f, "Enter 'YES' to continue or <ENTER> to cancel", "YES")
                 self.expect(f, "Press <ENTER> to continue...", "")
                 s.close()
                 sleep(3) #Power bricks have huge caps. Let power stabilize

    def power(self, on):
        #Sometimes we can connect, but get back nothing
        retries = 4
        while True:
            try:
                self._power(on)
                break
            except:
                continue

    def resetWithCustomFlags(self, flags=[]):
        return self.resetToHost(flags)

    def resetToHost(self, flags = []):
        self.power(0)
        sleep(3)
        self.power(1)

    def resetToNormal(self, flags = []):
        self.power(0)
        sleep(3)
        self.power(1)

    def add_argparse_options(parser):
        parser.add_argument("--apc-host",
                            help="APC IP Address/hostname",
                            default="192.168.10.2",
                            required=False)
        parser.add_argument("--apc-user",
                            help="APC IP username",
                            default="whiteblade",
                            required=False)
        parser.add_argument("--apc-pass",
                            help="APC IP username",
                            default="ash2wert",
                            required=False)
        parser.add_argument("--apc-outlet",
                            help="APC power outlet to use",
                            default=8,
                            required=False)



from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal
import os
import sys
import argparse
import rumboot_xrun
import rumboot
from parse import *

def cli():
    resets  = ResetSeqFactory("rumboot.resetseq")
    formats = ImageFormatDb("rumboot.images")
    chips   = ChipDb("rumboot.chips")

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-xrun {} - RumBoot X-Modem execution tool\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    helper = arghelper()

    helper.add_file_handling_opts(parser, True)
    helper.add_terminal_opts(parser)
    helper.add_resetseq_options(parser, resets)
    plus = parser.add_argument_group("Plusargs parser options",
        """
        rumboot-xrun can parse plusargs (similar to verilog simulator)
        and use them for runtime file uploads. This option is intended
        to be used for
        """)
    plus.add_argument('-A', '--plusargs', nargs='*')
    parser.add_argument("-R", "--rebuild",
                        help="Attempt to rebuild/update target before uploading",
                        action="store_true",
                        default=False,
                        required=False)

    parser.add_argument("-I", '--stdin',
                        help="Use stdin redirection to tty",
                        action="store_true",
                        default=False,
                        required=False)
    parser.add_argument('--replay-no-exit',
                        help="Do not exit on panics/rom returns when replaying logs (for batch analysis)",
                        action="store_true",
                        default=False,
                        required=False)
    if len(sys.argv) == 1:
        sys.argv = sys.argv + ["--help"]

    opts = parser.parse_args()

    #Open files, rebuild if needed
    opts.file[0], dumps = helper.process_files(opts.file[0], opts.rebuild)

    dump_path = os.path.dirname(__file__) + "/romdumps/"

    plusargs = {}
    if opts.plusargs:
        for a in opts.plusargs:
            ret = parse("+{}={}", a)
            if ret:
                plusargs[ret[0]] = ret[1]
                continue
            ret = parse("+{}", a)
            if ret:
                plusargs[ret[0]] = True

    c = helper.detect_chip_type(opts, chips, formats)
    if c == None:
        return 1

    print("Detected chip:    %s (%s)" % (c.name, c.part))
    if c == None:
        print("ERROR: Failed to auto-detect chip type")
        return 1
    if opts.baud == None:
        opts.baud = [ c.baudrate ]

    reset = resets[opts.reset[0]](opts)
    term = terminal(opts.port[0], opts.baud[0])
    term.set_chip(c)
    term.plusargs = plusargs
    term.xfer.xfers["edcl"].force_static_arp = opts.force_static_arp

    try:
        romdump = open(dump_path + c.romdump, "r")
        term.add_dumps({'rom' : romdump})
    except:
        pass

    if opts.log:
        term.logstream = opts.log

    print("Reset method:               %s" % (reset.name))
    print("Baudrate:                   %d bps" % int(opts.baud[0]))
    print("Port:                       %s" % opts.port[0])
    reset.resetToHost()
    term.add_binaries(opts.file)
    term.add_dumps(dumps)
    term.replay_till_the_end = opts.replay_no_exit
    if opts.edcl and c.edcl != None:
        term.xfer.selectTransport("edcl")
    print("Preferred data transport:   %s" % term.xfer.how)
    return term.loop(opts.stdin)



from parse import *
class base:
    formats = {}        
    hidden = False
    def __init__(self, term):
        self.term = term

    #executed on first term.loop()
    def on_start(self):
        pass

    def handle_line(self, line):
        for key,value in self.formats.items():
            pres = parse(value, line)
            if pres != None:
                retval = self.action(key, pres)
                if type(retval) is int:
                    return retval
                if retval == True:
                    return retval
        return False

    #True  - stop processing
    #False - go on processing (not processed)
    #Number - exit code
    def action(self, trigger, result):
        print("base:", self, trigger, result)
        return False



from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal
from rumboot.server import server
import os
import argparse
import rumboot_xrun
import rumboot
from parse import *

def cli():
    resets  = ResetSeqFactory("rumboot.resetseq")
    formats = ImageFormatDb("rumboot.images")
    chips   = ChipDb("rumboot.chips")

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-daemon {} - Collaborative board access daemon\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    helper = arghelper()
    helper.add_terminal_opts(parser)
    helper.add_file_handling_opts(parser)
    helper.add_resetseq_options(parser, resets)
    parser.add_argument("-L", "--listen",
                        help="Specify address:port to listen (default 0.0.0.0:10000)",
                        nargs=1, metavar=('listen'), default=["0.0.0.0:10000"],
                        required=False)
    opts = parser.parse_args()

    c = helper.detect_chip_type(opts, chips, formats)
    if c == None:
        c = chips["basis"]
        return 1
        
    print("Detected chip:    %s (%s)" % (c.name, c.part))
    if c == None:
        print("ERROR: Failed to auto-detect chip type")
        return 1
    if opts.baud == None:
        opts.baud = [ c.baudrate ]

    reset = resets[opts.reset[0]](opts)

    if opts.log:
        term.logstream = opts.log

    print("Reset method:     %s" % (reset.name))
    print("Baudrate:         %d bps" % int(opts.baud[0]))
    print("Serial Port:      %s" % opts.port[0])
    print("Listen address:   %s" % opts.listen)

    term = terminal(opts.port[0], opts.baud[0])
    srv = server(term, opts.listen[0])
    srv.set_reset_seq(reset)

    if "file" in opts:
        srv.preload_binaries(opts.file)

    return srv.loop()
    



import os
import sys
import argparse
import threading
import time

from parse import parse
from hexdump import hexdump

from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal
import rumboot_xrun
import rumboot

#TODO: Passthrough all missing methods to self.xfer
#TODO: Preload sio spl, when required
#TODO: Implement missing API in xfer.py for edcl and serial

class engine():
    def preload(self):
        self.reset()

    def __init__(self, target_chip=None, transport="xmodem", args=None):
        resets  = ResetSeqFactory("rumboot.resetseq")
        formats = ImageFormatDb("rumboot.images")
        chips   = ChipDb("rumboot.chips")
        c       = chips[target_chip] 

        parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                         description="Rumboot Scripting Engine script, using rumboot-tools {}\n".format(rumboot.__version__) +
                                        rumboot.__copyright__)
        helper = arghelper()

        helper.add_terminal_opts(parser)
        helper.add_resetseq_options(parser, resets)

        opts = parser.parse_args()

        dump_path = os.path.dirname(__file__) + "/romdumps/"

        helper.detect_terminal_options(opts, c)

        print("Target Chip:    %s (%s)" % (c.name, c.part))
        if c == None:
            raise Exception("ERROR: Failed to auto-detect chip type")

        if opts.baud == None:
            opts.baud = [ c.baudrate ]

        reset = resets[opts.reset[0]](opts)
        term = terminal(opts.port[0], opts.baud[0])
        term.set_chip(c)

        try:
            romdump = open(dump_path + c.romdump, "r")
            term.add_dumps({'rom' : romdump})
        except:
            pass

        if opts.log:
            term.logstream = opts.log

        print("Reset method:               %s" % (reset.name))
        print("Baudrate:                   %d bps" % int(opts.baud[0]))
        print("Port:                       %s" % opts.port[0])
        if opts.edcl and c.edcl != None:
            term.xfer.selectTransport("edcl")
        print("Preferred data transport:   %s" % term.xfer.how)
        print("--- --- --- --- --- --- --- --- --- ")
        self.terminal = term
        self.xfer = term.xfer

    def reset(self):
        self.reset.resetToHost()

    #TODO: stdout processing
    def run(self, spl, stdout=None):
        self.term.add_binaries(spl)
        self.terminal.add_binaries(spl)
        return self.terminal.loop(False,True)

    def read32(self, addr):
        return self.xfer.read32(addr)

    def write32(self, addr, data):
        return self.xfer.write32(addr, data)

    def read8(self, addr):
        return self.xfer.read8(addr)

    def write8(self, addr, data):
        return self.xfer.write8(addr,data)

    def read16(self, addr):
        return self.xfer.read16(addr)

    def write16(self, addr, data):
        return self.xfer.write16(addr,data)

    def read64(self, addr):
        return self.xfer.read64(addr)

    def write64(self, addr, data):
        return self.xfer.write64(addr,data)

    def read(self, addr, len):
        return self.xfer.read(addr, len)

    def write(self, addr, data):
        return self.xfer.write(addr, data)

    def dump(self, addr, len):
        data = self.read(addr, len)
        hexdump(data)

    def upload(self, file, addr):
        pass

    def download(self, file, addr, length):
        pass



from rumboot.chips.base import chipBase

class mm7705(chipBase):
    part="1888ТХ018"
    chip_id=1
    chip_rev=1
    skipsync=True
    spl_address=0x48000
    warning='''
    This chip has a RumBoot V1 bootloader. 
    RumBootV2 may be flashed onto SPI flash and executed from internal memory
    to provide additional functionality, e.g. xmodem loading
    Please be careful not wipe it! Recovery will be only possible via JTAG/EDCL
    '''
    welcome='host'
    baudrate=1000000
    memories = {
        "spi" : "rumboot-mm7705-PostProduction-updater-spiflash.bin"
    }
    edcl = [
    {   
        "name"   : "Greth 100Mbit #0", 
        "ip"     : "192.168.1.2", 
        "mac"    : "ec:17:66:00:00:02"
    },
    {   
        "name"   : "Greth 100Mbit #1", 
        "ip"     : "192.168.1.3", 
        "mac"    : "ec:17:66:00:00:03"
    },
    {   
        "name"   : "Greth 100Mbit #2", 
        "ip"     : "192.168.1.0", 
        "mac"    : "ec:17:66:00:00:00"
    },
    {   
        "name"   : "Greth 1Gbit #0", 
        "ip"     : "192.168.1.49", 
        "mac"    : "ec:17:66:77:05:01"
    },
    {   
        "name"   : "Greth 1Gbit #1", 
        "ip"     : "192.168.1.48", 
        "mac"    : "ec:17:66:77:05:00"
    }
    ]

    hacks = {
        "skipsync"   : True, # Doesn't Send U\r\n\r\n at the start.
    }

    romdump = "mm7705-v1.dmp"    
    
class basis(chipBase):
    part="1888ВС048"
    chip_id=3
    chip_rev=1
    welcome='host'
    baudrate=115200
    spl_address = 0x40000
    memories = {
        "i2c0-0x50": "rumboot-basis-PostProduction-updater-i2c0-0x50.bin",
        "i2c0-0x51": "rumboot-basis-PostProduction-updater-i2c0-0x51.bin", 
        "i2c0-0x52": "rumboot-basis-PostProduction-updater-i2c0-0x52.bin", 
        "i2c0-0x53": "rumboot-basis-PostProduction-updater-i2c0-0x53.bin",
        "spi0-gpio0_5-cs": "rumboot-basis-PostProduction-updater-spi0-gpio0_5-cs.bin",
        "spi0-internal-cs": "rumboot-basis-PostProduction-updater-spi0-internal-cs.bin",
        "spi1-internal-cs": "rumboot-basis-PostProduction-updater-spi1-internal-cs.bin",
    }
    flashrom = {
        "spi0-gpio0_5-cs": "rumboot-basis-PostProduction-serprog-spi0-gpio0_5-cs.bin",
        "spi0-internal-cs": "rumboot-basis-PostProduction-serprog-spi0-internal-cs.bin",
        "spi1-internal-cs": "rumboot-basis-PostProduction-serprog-spi1-internal-cs.bin",
    }
    romdump = "basis-v1.dmp"    

class oi10(chipBase):
    part="1888ВМ018(A)/1888ВМ01H4"
    chip_id=4
    chip_rev=1
    welcome='host'
    baudrate=115200
    memories = {
        "spi0-internal-cs": "rumboot-oi10-PostProduction-updater-spi-flash-0.bin",
        "nor":  "rumboot-oi10-PostProduction-updater-nor-mt150.04.bin",
        "nor-bootrom":  "rumboot-oi10-PostProduction-updater-nor-mt150.04-brom.bin"
    }
    stub = "rumboot-oi10-PostProduction-gdb-stub.bin"
    romdump = "oi10-v1.dmp"
    gdb = "powerpc-rcm-elf-gdb"
    spl_address = 0xC0000000
    edcl = [
    {   
        "name"   : "Greth #1", 
        "ip"     : "192.168.1.48", 
        "mac"    : "ec:17:66:0e:10:00"
    },
    {   
        "name"   : "Greth #2", 
        "ip"     : "192.168.1.49", 
        "mac"    : "ec:17:66:0e:10:01"
    },
    ]

class bbp3(chipBase):
    part="1888ВС058"
    chip_id=5
    baudrate=115200
    chip_rev=1
    welcome='host'
    baudrate=115200
    spl_address=0x8000
    memories = {
        "spi0-cs0": "rumboot-bbp3-PostProduction-updater-spi0-cs0.bin",
    }
    romdump = "bbp3-v1.dmp"    
    edcl = [
        {   
            "name"   : "GRETH", 
            "ip"     : "192.168.1.48", 
            "mac"    : "ec:17:66:ae:de:30"
        }
    ]

class nm6408(chipBase):
    part="1888ВС058"
    chip_id=6
    chip_rev=1
    welcome='host'
    baudrate=115200
    spl_address = 0x0
    warning='''
    This chip has a legacy ROM bootloader. 
    RumBoot is flashed onto SPI flash and executed from internal memory
    to provide additional functionality
    Please be careful not wipe it! Recovery will be only possible via JTAG
    '''


    #This chip's mac/ip are configurable via 4 boot pins
    #Let's handle it here
    def __populate_6408_edcl_params():
        edcl = []
        for i in range(0,16):
            greth = {   
                "name"   : "GRETH #" + str(i), 
                "ip"     : "192.168.1."+str(i), 
                "mac"    : "ec:17:66:64:08:"+str(i)
            }
            edcl.append(greth)
        return edcl

    edcl = __populate_6408_edcl_params()

    memories = {
        "spi0-cs0": "rumboot-nm6408-PostProduction-updater-spi0-cs0.bin",
    }

    hacks = {
        "skipsync"   : True, # Doesn't Send U\r\n\r\n at the start. All legacy stuff
        "silentRom"  : True, # Bootrom is totally silent
        "noxmodem"   : True,  # Chip lacks xmodem implementation
    }

class mb7707(chipBase):
    part="K1879XB1YA"
    chip_id=2
    chip_rev=1
    welcome='host'
    baudrate=38400
    spl_address = 0x00100000
    warning='''
    This chip has a legacy ROM bootloader. 
    RumBoot is flashed onto NAND flash and executed from internal memory
    to provide additional functionality
    Please be careful not wipe it! Recovery will be only possible via JTAG/EDCL
    '''
    memories = {

    }

    hacks = {
        "skipsync"   : True, # Doesn't Send U\r\n\r\n at the start. All legacy stuff
        "silentRom"  : True, # Bootrom is totally silent
        "noxmodem"   : True,  # Chip lacks xmodem implementation
    }

    edcl = [
        {   
            "name"   : "GRETH", 
            "ip"     : "192.168.1.0", 
            "mac"    : "0:0:5e:0:0:0"
        }
    ]



from rumboot.images.imageFormatBase import ImageFormatBase
import hashlib

class ImageFormatLegacyK1879XB1YA(ImageFormatBase):
    """
    16 bytes starting with 0x77077707 when empty, md5 of 4k of data otherwise
    16 bytes 0x0 when empty, md5 of 4k of data otherwise
    """
    name = "Legacy K1879XB1YA"
    MAGIC = 0x77077707
    format = [
        [16, "magic", "0x%x", "MD5/MAGIC"],
    ]


    def check(self):
        self.read_header()
        #Do we have a magic? 
        if (self.header["magic"] == self.MAGIC):
            self.endian = "little"
            return True
        #Or perhaps we're already signed?
        self.fd.seek(0)
        md5_orig = self.fd.read(16)
        k4 = self.fd.read(4*1024 - 16)
        md5 = hashlib.md5(k4)
        if md5_orig.hex() == md5.hexdigest():
            return True

        return False

    def wrap(self):
        return True

    def fix_checksums(self, calc_data=True):
        self.fd.seek(16)
        k4 = self.fd.read(4*1024 - 16)
        self.fd.seek(0)
        md5 = hashlib.md5(k4)
        self.fd.write(md5.digest())


    def get_chip_id(self):
        return 2

    def get_chip_rev(self):
        return 1    



from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal

import os
import argparse
import rumboot_xflash
import rumboot_packimage
from parse import *
import rumboot
import threading
import socket
from rumboot.server import redirector
import gdbgui.backend
import sys


class ipgdb(threading.Thread):
    gdbcmds = [ ]

    def configure(self, gdb, port, args):
        self.gdb = gdb
        self.port = port
        self.args = args
        target = "target extended-remote 127.0.0.1:" + str(self.port)
        self.add_command("set remotetimeout 10")
        self.add_command(target)

    def add_command(self, cmd):
        self.gdbcmds = self.gdbcmds + [ cmd ]

    def enable_debug(self):
        self.gdbcmds = [ "set debug remote 1" ] + self.gdbcmds

    def gui(self, gui):
        self.gui = gui

    def run(self):
        cmd = self.gdb + " " + self.args
        for c in self.gdbcmds:
            c = c.replace('"','\\"')
            cmd = cmd + " " + "-ex '" + c + "'"
        print("Invoking gdb: " + cmd)
        return os.system(cmd)

class ipgdbgui(ipgdb):
    def run(self):
        sys.argv = [ sys.argv[0] ]
        sys.argv = sys.argv + [ "-g" + self.gdb ]
        arg = "--gdb-args="
        for c in self.gdbcmds:
            arg = arg + "-ex \"" + c + "\" "
        arg = arg + self.args
        sys.argv = sys.argv + [ arg + "" ]            
        print(sys.argv)
        gdbgui.backend.main()


def cli():
    resets  = ResetSeqFactory("rumboot.resetseq")
    formats = ImageFormatDb("rumboot.images")
    chips   = ChipDb("rumboot.chips")

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-gdb {} - Debugger launcher tool\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    helper = arghelper();                                
    helper.add_terminal_opts(parser)
    parser.add_argument("-v", "--verbose", 
                        action='store_true',
                        default=False,
                        help="Print serial debug messages during preload phase"),
    parser.add_argument("-z", "--spl-path",
                        help="Path for SPL writers (Debug only)",
                        type=str,
                        required=False)
    parser.add_argument("-L", "--load",
                        help="Load binary on start",
                        action='store_true',
                        default=False,
                        required=False)
    parser.add_argument("-f", "--file",
                        help="Application ELF file to debug",
                        type=str,
                        required=False)
    parser.add_argument("-x", "--exec",
                        help="Execute supplied binary on start (Implies --load)",
                        action='store_true',
                        default=False,
                        required=False)
    parser.add_argument("-R", "--rebuild",
                        help="Attempt to rebuild binary",
                        action='store_true',
                        default=False,
                        required=False)
    parser.add_argument("--debug-remote",
                        help="Send 'set debug remote 1' to gdb for debugging",
                        action='store_true',
                        default=False,
                        required=False)
    parser.add_argument("--debug-serial",
                        help="Send 'set debug serial 1' to gdb for debugging",
                        action='store_true',
                        default=False,
                        required=False)
    parser.add_argument("-g", "--gdb-path",
                        help="Path to flashrom binary",
                        type=str,
                        required=False)
    parser.add_argument("-G", "--gdb-gui",
                        help="Start GDB gui",
                        action='store_true',
                        default=False,
                        required=False)
    parser.add_argument("-c", "--chip_id",
                    help="Chip Id (numeric or name)",
                    required=True)
    parser.add_argument('remaining', nargs=argparse.REMAINDER, default=[], help="Extra gdb arguments (e.g. filename)")


    helper.add_resetseq_options(parser, resets)

    opts = parser.parse_args()
    if opts.remaining and opts.remaining[0] == "--":
        opts.remaining = opts.remaining[1:]
    gdb_args = " ".join(opts.remaining)

    c = chips[opts.chip_id]
    if (c == None):
        return 1;
 
    helper.detect_terminal_options(opts, c)
    print("Detected chip:    %s (%s)" % (c.name, c.part))

    if c.warning != None:
        print("    --- WARNING ---")
        print(c.warning)        
        print("    --- WARNING ---")
        print("")

    spl_path = os.path.dirname(__file__) + "/tools/"

    if opts.baud == None:
        opts.baud = [ c.baudrate ]

    reset = resets[opts.reset[0]](opts)
    term = terminal(opts.port[0], opts.baud[0])
    term.verbose = opts.verbose
    term.set_chip(c)
    if opts.log:
        term.logstream = opts.log
    
    try:
        spl = c.stub
    except:
        print("ERROR: Target chip (%s) doesn't have a gdb stub" % (c.name))
        return 1

    if opts.spl_path != None:
        spl_path = opts.spl_path
        print("SPL               %s" % (spl_path + c.stub))

    spl = spl_path + c.stub

    if opts.gdb_path:
        gdb = opts.gdb_path
    else:
        gdb = c.gdb

    print(opts.rebuild, opts.file)
    if opts.rebuild and opts.file != None:
        if 0!=os.system("cmake --build . --target " + opts.file):
            os.exit(1)

    print("Reset method:     %s" % (reset.name))
    print("Baudrate:         %d bps" % int(opts.baud[0]))
    print("Port:             %s" % opts.port[0])
    print("GDB:              %s" % gdb)

    if opts.edcl and c.edcl != None:
        term.xfer.selectTransport("edcl")

    reset.resetToHost()
    term.add_binaries(spl)

    term.loop(break_after_uploads=True)
    while True:
        s = term.ser.readline()
        if opts.verbose:
            print(s)
        if s.decode("utf-8", errors="replace").find("STUB READY!") != -1:
            break
    print("gdb-stub is ready for operations, starting gdb..")

    # If we're here, we'll need to setup socket redirector
    port=20000

    while True:
        try:
            print("Trying port %d" % port)
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.bind(("127.0.0.1", port))
            server.listen()            
            break
        except:
            port = port + 1
            if port > 30000:
                print("Weird shit going on, breaking")
                break

    #Fire up flashrom in background

    if opts.gdb_gui:
        flr = ipgdbgui()
    else:
        flr = ipgdb()

    if opts.debug_remote:
        flr.add_command("set debug remote 1")

    if opts.debug_serial:
        flr.add_command("set debug serial 1")

    if opts.file != None: 
        gdb_args = gdb_args + opts.file

    flr.configure(gdb, port, gdb_args)
    if opts.load or opts.exec:
        flr.add_command("load")

    if opts.exec:
        flr.add_command("continue")

    flr.start()

    #Accept connection
    connection, client_address = server.accept()
    connection.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    rd = redirector()
    rd.configure(term.ser, connection)
    rd.start()

    return flr.join()
    






import sys
import os
import tempfile
class basic:
    name = "basic"
    align = 1
    fd = None

    def __init__(self, outfile, align = 1):
        align = int(align)
        self.align = align
        self.outfile = outfile
        self.fd = open(outfile + ".appending", "wb+")

    def append(self, infile):
        opos = self.fd.tell()
        while opos % self.align > 0:
            self.fd.write(b'\00')
            opos = opos + 1
        infile.seek(0, os.SEEK_END)
        size = infile.tell()
        infile.seek(0)
        data = infile.read()
        self.fd.write(data)
        infile.close()

    def close(self):
        self.fd.close()
        #Avoid exception on windows 
        if os.path.exists(self.outfile):
            os.remove(self.outfile)
        os.rename(self.outfile + ".appending", self.outfile)

    def describe(self):
        return self.name + "/" + str(self.align) + " bytes"


class physmap(basic):
    name = "physmap"
    def __init__(self, outfile, align):
        super().__init__(outfile, 8)

class SD(basic):
    name = "SD"
    def __init__(self, outfile, align):
        super().__init__(outfile, 512)

class ini(basic):
    name = "ini"
    first = True
    def __init__(self, outfile, align):
        super().__init__(outfile)
    
    def append(self, infile):
        super().append(infile)

        if not self.first:
            self.fd.write(b'\00')

        self.first = False




from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal

import os
import argparse
import rumboot_xflash
import rumboot_packimage
from parse import *
import rumboot

def cli():
    resets  = ResetSeqFactory("rumboot.resetseq")
    formats = ImageFormatDb("rumboot.images")
    chips   = ChipDb("rumboot.chips")

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-xflash {} - RumBoot X-Modem firmware update tool\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    helper = arghelper();                                
    helper.add_file_handling_opts(parser, True)
    helper.add_terminal_opts(parser)
    parser.add_argument("-v", "--verbose", 
                        action='store_true',
                        help="Print serial debug messages during update"),
    parser.add_argument("-m", "--memory",
                        help="Memory program. Help for a list of memories",
                        nargs=1, metavar=('memory'),
                        default="help",
                        required=True)                        
    parser.add_argument("-z", "--spl-path",
                        help="Path for SPL writers (Debug only)",
                        type=str,
                        required=False)

    helper.add_resetseq_options(parser, resets)

    opts = parser.parse_args()

    c = helper.detect_chip_type(opts, chips, formats)
    if (c == None):
        return 1;

    mem = opts.memory[0]
    if not mem:
        return 1

    if mem == "help":
        for mem,spl in c.memories.items():
            print("Memory %16s: %s" % (mem, spl))
        return 1


    #Open files, rebuild if needed
    opts.file[0], dumps = helper.process_files(opts.file[0], False)

    print("Detected chip:    %s (%s)" % (c.name, c.part))
    if c.warning != None:
        print("    --- WARNING ---")
        print(c.warning)        
        print("    --- WARNING ---")
        print("")

    spl_path = os.path.dirname(__file__) + "/spl-tools/"

    if opts.baud == None:
        opts.baud = [ c.baudrate ]

    reset = resets[opts.reset[0]](opts)
    term = terminal(opts.port[0], opts.baud[0])
    term.set_chip(c)
    if opts.log:
        term.logstream = opts.log
    
    term.verbose = opts.verbose
    term.xfer.xfers["edcl"].force_static_arp = opts.force_static_arp

    try:
        spl = c.memories[mem]
    except:
        print("ERROR: Target chip (%s) doesn't have memory '%s'. Run with -m help for a list" % (c.name, mem))
        return 1

    if opts.spl_path != None:
        spl_path = opts.spl_path
        print("SPL                              %s" % (spl_path + c.memories[mem]))

    spl = spl_path + c.memories[mem]

    print("Reset method:                    %s" % (reset.name))
    print("Baudrate:                        %d bps" % int(opts.baud[0]))
    print("Port:                            %s" % opts.port[0])

    reset.resetToHost()
    term.add_binaries(spl)
    term.add_binaries(opts.file[0][0])
    if opts.edcl and c.edcl != None:
        term.xfer.selectTransport("edcl")
    print("Preferred data transport:        %s" % term.xfer.how)

    reset.resetToHost()
    term.loop()



class chipBase:
    name = "whatever" # Will be replaced by class name during load!
    part="part"
    chip_id=0
    chip_rev=0
    gdb = "gdb"
    baudrate=115200
    warning=None
    memories = {}
    dumps = {}
    hacks = {
        "skipsync"   : False, # Doesn't Send U\r\n\r\n at the start. All legacy stuff
        "edclArpBug" : False, # EDCL doesn't have a valid IP, needs a static ARP entry (mb7707)
        "silentRom"  : False, # Bootrom is totally silent
        "noxmodem"   : False,  # Chip lacks xmodem implementation
        "noedcl"     : False  # Chip lacks edcl implementation
    }

    edcl = None
    




import subprocess
import os
import platform
import time

import netifaces as ni
from netaddr import IPNetwork, IPAddress
from parse import parse

from rumboot.edcl import edcl
from rumboot.chipDb import ChipDb


class edclmanager(object):
    edcl = None
    verbose = False
    chips = ChipDb("rumboot.chips")
    warned = []
    force_static_arp = False

    def __init__(self):
        self.edcl = edcl()

    def check_reachability(self, ip):
        for i in ni.interfaces():
            if not ni.AF_INET in ni.ifaddresses(i):
                continue
            for addr in ni.ifaddresses(i)[ni.AF_INET]:
                if IPAddress(ip) in IPNetwork(addr['addr'] + "/" +addr["netmask"]):
                    return True
        return False

    def buggy_ip(self, params):
        arr = params["ip"].split(".")
        return int(arr[3]) == 0
            
    def __getattribute__(self, key):
        try:
            return object.__getattribute__(self, key)
        except:
            return object.__getattribute__(self.edcl, key)

    def try_connect(self, chip, params):
        if self.buggy_ip(params) or self.force_static_arp:
            #FIXME: This logic sucks. 
            #TODO: Implement a better solution to find non-used IP on the subnet
            params["ip"] = params["ip"].replace(".0", ".76")
            self.prepareStaticARP(chip, params)
        if not self.check_reachability(params["ip"]) and not params["ip"] in self.warned:
            print("WARNING: IP address %s is not in your subnet." % params["ip"])
            print("WARNING: Please check network interface settings.")
            self.warned.append(params["ip"])
            return None

        if self.verbose:
            print("Testing: %s (%s) IP %s" % (chip.__name__, params["name"], params["ip"]))
        try: 
            ret = self.edcl.connect(params["ip"])
        except:
            return None
        if ret:
            print("Connected: %s (%s)" % (chip.__name__, params["name"]))
            return chip

    def probe(self, chip):
        for params in chip.edcl:
            if None != self.try_connect(chip, params):
                return params
        return None

    def scan(self, match=None):
        ret = []
        if type(match) == int or type(match) == str:
            match = self.chips[match]

        if match == None:
            chips = self.chips
        else:
            chips = { match }
        for c in chips:
            tmp = self.probe(c)
            if tmp != None:
                db = {
                    "chip" : c,
                    "params": tmp
                }
                ret.append(db)
        return ret

    def test_chip(self, chip):
        pass

    def showArpWarning(self, cmd):
        print("WARNING: Trying to add ARP record via 'sudo'/'runas'.")
        print("WARNING: If that doesn't work please run manually: %s" % cmd)

    def dropARP(self, chip, params):
        if platform.system() == 'Linux':
            cmd = "sudo arp -d %s" % params["ip"]
        elif platform.system() == 'Windows': 
            cmd = "runas /user:Administrator arp /d %s" % params["ip"]
        else:
            raise("FATAL: Don't know how to add static ARP on %s", platform.system())
        os.system(cmd)
        self.showArpWarning(cmd)

    def addARP(self, chip, params):
        if platform.system() == 'Linux':
            cmd = "sudo arp -s %s %s" % (params["ip"], params["mac"])
        elif platform.system() == 'Windows': 
            cmd = "runas /user:Administrator arp /s %s %s" % (params["ip"], params["mac"].upper().replace(":","-"))
        else:
            raise("FATAL: Don't know how to add static ARP on %s", platform.system())
        os.system(cmd)
        self.showArpWarning(cmd)
        
    def listARP(self):
        ret = {}
        if platform.system() == 'Linux':
            cmd = "/usr/sbin/arp -n"
        elif platform.system() == 'Windows': 
            cmd = "arp -a"
        else:
            raise("FATAL: I don't know how to query ARP on %s", platform.system())
        out = subprocess.Popen(cmd.split(" "), 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE)
        stdout,stderr = out.communicate()
        if platform.system() == 'Linux':
            for l in stdout.decode().split("\n"):
                result = parse("{:<} ether {:>} {:>} {:>}", l)
                if result == None:
                    continue
                ret[result[0]]=result[1]
        elif  platform.system() == 'Windows': 
            for l in stdout.decode().split("\n"):
                result = parse("{:<} {:>} {:>}", l)
                if result == None:
                    continue
                ret[result[0]]=result[1].replace("-",":")
        return ret

    def getARP(self, ip):
        tbl = self.listARP()
        if ip in tbl:
            return tbl[ip]

    def prepareStaticARP(self, chip, params):
        if self.getARP(params["ip"]) and not self.force_static_arp:
            print("Static ARP record already exists, good")
            return
            
        if self.force_static_arp:
            self.dropARP(chip, params)
        self.addARP(chip, params)

    def smartupload(self, address, image, callback = None):
        self.edcl.send_from_file(address + 4, image, callback, 4)
        self.edcl.send_from_file(address    , image, None, 0, 4)

    def connect(self, chip):
        for i in range(3):
            ret = self.scan(chip)
            if ret != []:
                return ret
    


import argparse
import os
import sys

try:
    import yaml
except:
    yaml = None

class arghelper():
    configs = [
        "~/.rumboot.yaml", 
        "/etc/rumboot.yaml", 
        os.path.dirname(__file__) + "/../rumboot.yaml"
        ]
    config = None
    default_port = "/dev/ttyUSB0"

    def get_default_port(self, chip):
        try:
            ret = self.config['xrun']['chips'][chip.__name__]["port"]
            return ret
        except:
            pass
        
        try: 
            ret = self.config['xrun']['defaults']["port"]
            return ret
        except:
            return self.default_port

    def get_default_baud(self, chip):
        try:
            ret = self.config['xrun']['chips'][chip.name]["baudrate"]
            return ret
        except:
            if chip != None:
                return chip.baudrate
            return 115200

    def __init__(self):
        if yaml == None:
            print("[W] No yaml, disabling config files. Install with 'pip install yaml' to fix")
            return
        for conf in self.configs:
            conf = os.path.expanduser(conf)
            if (os.path.exists(conf)):
                self.load_config(conf)
                break

    def load_config(self, conf):
        print("[!] Using configuration file: " + conf)
        with open(conf, 'r') as stream:
            try:
                self.config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)

    def process_files(self, files, rebuild):
        ret = files
        dumps = { }
        for k,f in enumerate(files):
            target = f.replace(".bin", ".all")
            dmp = f.replace(".bin", ".dmp")
            if rebuild:
                bres = os.system("cmake --build . --target %s" % target)
                if (bres != 0):
                    sys.exit(bres)

            ret[k] = open(f, "rb")

            try:
                dumps[f] = open(dmp, "r")
            except:
                print("WARN: Missing dump file: " + dmp + ". Detailed traces will be unavailable")
                pass

        return ret, dumps

    def add_file_handling_opts(self, parser, need_file=False):
        group = parser.add_argument_group('File Handling')
        parser.add_argument('-f','--file',action='append',nargs=1,
                        type=str,
                        help="Image file (may be specified multiple times)")
        group.add_argument("-c", "--chip_id",
                        help="Override chip id (by name or chip_id)",
                        nargs=1, metavar=('chip_id'),
                        required=False)            
    def add_terminal_opts(self, parser):
        group = parser.add_argument_group('Connection Settings')
        group.add_argument("-l", "--log",
                        help="Log terminal output to file",
                        type=argparse.FileType("w+"),
                        required=False)
        group.add_argument("-p", "--port",
                        help="Serial port to use",
                        nargs=1, metavar=('port'),
                        required=False),
        group.add_argument("-b", "--baud",
                        help="Serial line speed",
                        type=int,
                        nargs=1, metavar=('speed'),
                        required=False)
        group.add_argument("-e", '--edcl',
                        help="Use edcl for data uploads (when possible)",
                        action="store_true",
                        default=False,
                        required=False)
        group.add_argument('--force-static-arp',
                        help="Always add static ARP entries",
                        action="store_true",
                        default=False,
                        required=False)


    def add_resetseq_options(self, parser, rfactory):
        group = parser.add_argument_group('Reset Sequence options', "These options control how the target board will be reset")
        resetlist = ""
        for r in rfactory.classes:
            resetlist = resetlist + " " + r
        resetlist = resetlist.strip()
        group.add_argument("-r", "--reset",
                        help="Reset sequence to use (%s)" % resetlist,
                        nargs=1, metavar=('method'),
                        default="none",
                        required=False)
        for name, sequence in rfactory.classes.items():
            if hasattr(sequence, "add_argparse_options"):
                g = parser.add_argument_group(name + " reset sequence options")
                sequence.add_argparse_options(g)
        
    def detect_terminal_options(self, opts, chip):
        # Now, let's update default options, if needed
        if opts.port == None:
            opts.port = [ self.get_default_port(chip) ]
        if opts.baud == None:
            opts.baud = [ self.get_default_baud(chip) ]

    def detect_chip_type(self, opts, chips, formats):
        c = None
        try:
            t = formats.guess(opts.file[0][0])
            c = chips[t.get_chip_id()]
        except:
            pass

        if opts.chip_id != None:
            c = chips[opts.chip_id[0]]

        if c == None:
            print("ERROR: Failed to auto-detect chip type")
            print("HINT: Specify a file (-f) or set chip id manually (-c)")

        self.detect_terminal_options(opts, c)
        return c



from rumboot.images.imageFormatBase import ImageFormatBase

class ImageFormatV2(ImageFormatBase):
    """
    This class works with version 2.0 images.

    struct __attribute__((packed)) rumboot_bootheader {
    uint32_t magic; /* 0xb0ldface */
    uint8_t  version;
    uint8_t  reserved;
    uint8_t  chip_id;
    uint8_t  chip_rev;
    uint32_t data_crc32;
    uint32_t datalen;
    uint32_t entry_point[11];
    uint32_t header_crc32;
    const struct rumboot_bootsource *device;
    char     data[];
};

    """
    name = "RumBootV2"
    MAGIC = 0xb01dface
    VERSION = 2
    format = [
        [4, "magic", "0x%x", "Magic"],
        [1, "version", "0x%x", "Header Version"],
        [1, "reserved", "0x%x", "Reserved"],
        [1, "chip_id", "0x%x", "Chip ID"],
        [1, "chip_rev", "0x%x", "Chip Revision"],
        [4, "data_crc32", "0x%x", "Data CRC32"],
        [4, "data_length", "%d", "Data Length"],
        [4, "entry0",  "0x%x", "Entry Point[0]"],
        [4, "entry1",  "0x%x", "Entry Point[1]"],
        [4, "entry2",  "0x%x", "Entry Point[2]"],
        [4, "entry3",  "0x%x", "Entry Point[3]"],
        [4, "entry4",  "0x%x", "Entry Point[4]"],
        [4, "entry5",  "0x%x", "Entry Point[5]"],
        [4, "entry6",  "0x%x", "Entry Point[6]"],
        [4, "entry7",  "0x%x", "Entry Point[7]"],
        [4, "entry8",  "0x%x", "Entry Point[8]"],
        [4, "entry9",  "0x%x", "Entry Point[9]"],
        [4, "header_crc32", "0x%x", "Header CRC32"],
        [4, "bootsource", "0x%x", ""],
    ]

    header = {}

    def dump_header(self, raw=False, format=False):
        #Hide all unused entry points
        for i in range(1, 10):
            key = "entry" + str(i)
            self.hide_field(key)
        #Dump fields
        return super().dump_header(raw, format)

    def check(self):
        if (super().check()):
            return self.header["version"] == self.VERSION
        return False

    def __init__(self, inFile):
        super().__init__(inFile)
    
    def wrap(self):
        super().wrap()
        self.header["version"] = 2
        self.write_header()
        return True

    def get_chip_id(self):
        return self.header['chip_id']

    def get_chip_rev(self):
        return self.header['chip_rev']



import os
import binascii
import sys

class ImageFormatBase:
    cur_pos = 0
    file_size = 0
    header_size = 64
    data_offset = 68
    endian = "little"
    debug = False
    logger = False;
    tblfmt = "%-24s"
    header_crc32 = 0
    data_crc32   = 0
    data_length  = 0
    MAGIC = -1
    header = {}

    # These field are common and obligatory
    format = [
        [4, "magic", "%x", "Magic"],
        [4, "data_crc32", "0x%x", "Data CRC32"],
        [4, "data_length", "%d", "Data Length"],
        [4, "header_crc32", "0x%x", "Header CRC32"],
    ]

    hidden = {}

    name = "Base"
    def __init__(self, inFile):
        if (type(inFile) is str):
            self.fd = open(inFile, 'r+b')
        else:
            self.fd = inFile
        self.read_file_size()
        self.header_size = 66

    def wrap(self):
        self.fd.seek(0, os.SEEK_SET)
        tmp = self.fd.read(self.file_size)
        for n,v in enumerate(self.header):
            self.header[v] = 0
        self.header["magic"] = self.MAGIC
        self.write_header()
        self.fd.seek(self.get_header_length(), os.SEEK_SET)
        self.fd.write(tmp)
        self.read_file_size()
        self.read_header()
        
    def read_file_size(self):
        self.fd.seek(0, os.SEEK_END)
        self.file_size = self.fd.tell()
        self.fd.seek(0, os.SEEK_SET)

    def field_exists(self, name):
        for f in self.format:
            if (f[1] == name):
                return True
        return False

    def hide_field(self, field):
        self.hidden[field] = True

    def dump_field(self, field, format=False, comment=False, raw=False, nocheck=False):

        if (format == False):
            format = self.format

        if(type(field) is list):
            f = field
        else:
            for f in self.format:
                if (f[1]==field):
                    break;

        if raw:
            label = f[1]
        else:
            label = f[3]

        if (comment):
            print((self.tblfmt + f[2] + " [%s]") % (label + ":", self.header[f[1]], comment))
        else:
            print((self.tblfmt + f[2]) % (label + ":", self.header[f[1]]))

    def dump_header(self, raw=False, format=False):
        #Hide fields that need comments from printout
        valid = True
        self.hide_field("header_crc32")
        self.hide_field("data_crc32")
        self.hide_field("data_length")

        if (format == False):
            format = self.format
        print("=== %s Header Information ===" % self.name)
        print((self.tblfmt + "%s") % ("Endianess:", self.endian))
        for f in format:
            show = True
            if (f[1] in self.hidden):
                show = False

            if (show and f[3]):
                self.dump_field(f, raw=raw)

        if (self.header["header_crc32"] == self.header_crc32):
            hmatch = "Valid"
        else:
            hmatch = "Invalid, expected: 0x{:X}".format(self.header_crc32)
            valid = False

        if (hmatch != "Valid"):
            dmatch = "Skipped, invalid header checksum"
        elif (self.header["data_crc32"] == self.data_crc32):
            dmatch = "Valid"
        else:
            dmatch = "Invalid, expected: {:X}".format(self.data_crc32)
            valid = False

        actual = False
        if (self.file_size - self.get_header_length() != self.header["data_length"]):
            actual = "Actual data length {}".format(self.file_size - self.get_header_length())

        if (self.field_exists("data_length")):
            self.dump_field("data_length",  False, actual, raw)
        if (self.field_exists("header_crc32")):
            self.dump_field("header_crc32", False, hmatch, raw)
        if (self.field_exists("data_crc32")):
            self.dump_field("data_crc32",   False, dmatch, raw)
        return valid

    def read_header(self):
        offset = 0
        for f in self.format:
            self.header[f[1]] = self.read_element(offset, f[0])
            offset = offset + f[0]
        self.header_crc32 = self.crc32(0, self.get_header_checksum_length())

        if (self.header["data_length"] == 0):
            self.data_length = self.file_size - self.get_header_length()
        else:
            self.data_length = self.header["data_length"]
            #Only compute data crc32 for a valid header checksum
        if (self.header["header_crc32"] == self.header_crc32):
            self.data_crc32 = self.crc32(self.get_header_length(), self.get_header_length() + self.data_length)

    def write_header(self):
        offset = 0
        for f in self.format:
            self.write_element(offset, self.header[f[1]], f[0])
            offset = offset + f[0]

    def fix_length(self):
        self.header["data_length"] = self.file_size - self.get_header_length()
        self.data_length = self.header["data_length"]

    def fix_checksums(self, calc_data = True):
        self.header["data_length"] = self.data_length
        if calc_data:
            self.header["data_crc32"] = self.crc32(self.get_header_length(), self.get_header_length() + self.data_length)
        self.write_header()
        self.header["header_crc32"] = self.crc32(0, self.get_header_checksum_length())
        self.write_header()
        self.read_header()

    def get_header_length(self):
        offset = 0
        for f in self.format:
            offset = offset + f[0]
        return offset

    def get_header_checksum_length(self):
        offset = 0
        for f in self.format:
            if (f[1]=="header_crc32"):
                break;
            offset = offset + f[0]
        return offset

    def read_element(self, offset, len, whence = os.SEEK_SET):
        self.fd.seek(offset, whence)
        tmp = self.fd.read(len)
        tmp = int.from_bytes(tmp, self.endian)
        return tmp

    def write_element(self, offset, value, len, whence = os.SEEK_SET):
        self.fd.seek(offset, whence)
        self.fd.write(bytearray(value.to_bytes(len, self.endian)))
        return value

    def read64(self, offset, whence = os.SEEK_SET):
        return self.read_element(offset, 8, whence)

    def read32(self, offset, whence = os.SEEK_SET):
        return self.read_element(offset, 4, whence)

    def read16(self, offset, whence = os.SEEK_SET):
        return self.read_element(offset, 2, whence)

    def read8(self, offset, whence = os.SEEK_SET):
        return self.read_element(offset, 1, whence)

    def write32(self, offset, value, whence = os.SEEK_SET):
        return self.write_element(offset, value, 4, whence)

    def write16(self, offset, value, whence = os.SEEK_SET):
        return self.write_element(offset, value, 2, whence)

    def write8(self, offset, value, whence = os.SEEK_SET):
        return self.write_element(offset, value, 1, whence)

    def crc32(self, from_byte, to_byte=-1):

        self.fd.seek(from_byte, os.SEEK_SET)
        if (to_byte == -1):
            to_byte = self.file_size
        data = self.fd.read(to_byte-from_byte)
        crc32 = (binascii.crc32(data) & 0xffffffff)
        return crc32

    def swap_endian(self):
        if (self.endian == "big"):
                self.endian = "little"
        if (self.endian == "little"):
                self.endian = "big"

    def check(self):
        self.read_header()
        if (self.header["magic"] != self.MAGIC):
            self.endian = "big"
            self.read_header()
            if (self.header["magic"] != self.MAGIC):
                return False
        self.read_header()
        return True

    def set(self, key, value):
        if not key in self.header:
            print("No such key in header: " + key)
        self.header[key] = int(value,16);
        self.write_header()
        self.read_header()

    def get(self, key):
        return self.header[key]

    def add_zeroes(self, count):
        self.fd.seek(0, os.SEEK_END)
        len = int(count)
        self.file_size += len
        while len > 0:
            self.fd.write(bytearray([0]))
            len = len - 1

    def align(self, align):
        self.fd.seek(0, os.SEEK_END)
        align = int(align)
        if not self.file_size % align:
            return
        len = align - (self.file_size % align)
        self.file_size += len 
        while len > 0:
            self.fd.write(bytearray([0]))
            len = len - 1

    def get_chip_id(self):
        return 0

    def get_chip_rev(self):
        return 0


from rumboot.classLoader import classLoader

class LayoutFactory(classLoader):
    def __getitem__(self, key):
        if key in self.classes:
            return self.classes[key]
        return self.classes["basic"]


from rumboot.ops.base import base
from rumboot.ops.xfer import basic_uploader
import tqdm
import rumboot.xfer


class edcl_generic_uploader(basic_uploader):
    formats = {
        "mm7705": "\x00boot: Waiting for a valid image @ {:x}"
    }

    def action(self, trigger, result):
        tp = self.term.formats.guess(self.term.runlist[0])
        if not tp or tp.name != "RumBootV1":
            return False

        print("WARNING: Bootloader doesn't support xmodem, forcing edcl upload")
        old = self.term.xfer.how
        self.term.xfer.selectTransport("edcl")
        if not self.term.xfer.push(result[0]):
            return 1
        self.term.xfer.selectTransport(old)
        return True


class dumb_chips_uploader(basic_uploader):
    def on_start(self):
        #HACK: Old chips don't print anything via uart
        #HACK: edcl is the only way to bring 'em up
        #HACK: We do the lucky guess by looking at the image header
        def prg(total_bytes, position, increment):
                self.term.progress.update(increment)     

        tp = self.term.formats.guess(self.term.runlist[0])
        if self.term.hack("silentRom"):
            print("WARNING: Bootloader doesn't support xmodem, forcing initial edcl upload")
            old = self.term.xfer.how
            self.term.xfer.selectTransport("edcl")
            if not self.term.xfer.push(self.term.chip.spl_address):
                return 1
            self.term.xfer.selectTransport(old)

        if not tp:
            return True

        if tp.name == "Legacy K1879XB1YA":
            self.term.xfer.write32(0x00100010, 0x00100014)
            return True
        if tp.name == "NM6408 (Legacy)":
            self.term.xfer.write32(0x30000, 0x1)
            return True


from rumboot.chipDb import ChipDb
from rumboot.ImageFormatDb import ImageFormatDb
from rumboot.resetSeq import ResetSeqFactory
from rumboot.cmdline import arghelper
from rumboot.terminal import terminal
import os
import sys
import argparse
import rumboot_xrun
import rumboot
from parse import *
import atexit
import threading
import time


# TODO & questions:
# - Timeout handling in tests
# - Is atexit a good way to call the actual testing?
# - Decorator & class instance registration methods
# - Helper functions to add tests from directory
# - GUI integration

class RumbootTestBase(threading.Thread):
    terminal = None

    def prepare(self, terminal):
        self.terminal = terminal

    def run(self):
        return self.execute(self.terminal)


class RumbootTestFacility():
    runlist = []
    def __init__(self):
        pass
    
    def register(self, test):
        self.runlist.append(test())

    def run(self, terminal, reset):
        for r in self.runlist:
            reset.resetToHost()
            r.prepare(terminal)
            r.start()
            print("test done", r.join())


__g_facility = RumbootTestFacility()


def RumbootTest(arg):
    print("Registering test", arg)
    __g_facility.register(arg)


def intialize_testing(target_chip=None):
    resets  = ResetSeqFactory("rumboot.resetseq")
    formats = ImageFormatDb("rumboot.images")
    chips   = ChipDb("rumboot.chips")
    c       = chips[target_chip] 

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="rumboot-factorytest {} - RumBoot Board Testing Framework\n".format(rumboot.__version__) +
                                    rumboot.__copyright__)
    helper = arghelper()

    helper.add_terminal_opts(parser)
    helper.add_resetseq_options(parser, resets)

    opts = parser.parse_args()

    dump_path = os.path.dirname(__file__) + "/romdumps/"

    helper.detect_terminal_options(opts, c)

    print("Detected chip:    %s (%s)" % (c.name, c.part))
    if c == None:
        print("ERROR: Failed to auto-detect chip type")
        return 1
    if opts.baud == None:
        opts.baud = [ c.baudrate ]

    reset = resets[opts.reset[0]](opts)
    term = terminal(opts.port[0], opts.baud[0])
    term.set_chip(c)

    try:
        romdump = open(dump_path + c.romdump, "r")
        term.add_dumps({'rom' : romdump})
    except:
        pass

    if opts.log:
        term.logstream = opts.log

    print("Reset method:               %s" % (reset.name))
    print("Baudrate:                   %d bps" % int(opts.baud[0]))
    print("Port:                       %s" % opts.port[0])
    if opts.edcl and c.edcl != None:
        term.xfer.selectTransport("edcl")
    print("Preferred data transport:   %s" % term.xfer.how)
    print("--- --- --- --- --- --- ")

    return term, reset



@atexit.register
def execute():
    term, reset = intialize_testing("mm7705")
    __g_facility.run(term, reset)



from rumboot.classLoader import classLoader

class ImageFormatDb(classLoader):
    def guess(self, ifile):
        for n,f in self.classes.items():
            tmp = f(ifile)
            if tmp.check():
                return tmp
        return False

    def wrap(self, ifile, fmt):
        for n,f in self.classes.items():
            if fmt in n:
                tmp = f(ifile)
                if tmp.wrap():
                    return tmp
        return False


from rumboot.images.imageFormatBase import ImageFormatBase
import os


class ImageFormatLegacyNM6408(ImageFormatBase):
    MAGIC =  0x12345678
    name = "NM6408 (Legacy)"
    format = [
        [4, "magic", "0x%x", "Magic"],
        [4, "data_length", "%d", "Data Length"],
    ]

    def __init__(self, inFile):
        super().__init__(inFile)
        self.header_size = 8

    def get_header_length(self):
        return 8

    def dump_header(self, raw=False, format=False):
        valid = super().dump_header(raw, format)
        crc32 = self.read32(-4, os.SEEK_END)
        self.header["data_crc32"] = crc32
        if (crc32 == self.data_crc32):
            hmatch = "Valid"
        else:
            hmatch = "Invalid, expected: 0x{:X}".format(self.data_crc32)
            valid = False
        self.dump_field([4, "data_crc32", "0x%x", "Data CRC32"],  True, hmatch)
        return valid

    def read_header(self):
        offset = 0
        for f in self.format:
            self.header[f[1]] = self.read_element(offset, f[0])
            offset = offset + f[0]

        if self.header['magic'] != self.MAGIC:
            return

        if (self.header["data_length"] == 0):
            self.data_length = self.file_size - self.get_header_length()
        else:
            self.data_length = self.header["data_length"]
            #Only compute data crc32 for a valid header checksum

        self.data_crc32 = self.crc32(self.get_header_length(), self.get_header_length() + self.data_length - 4)
        
             
    def fix_length(self):
        self.header["data_length"] = (self.file_size - self.get_header_length()) & 0xFFFFFF
        self.data_length = self.header["data_length"]
        self.fix_checksums()

    def fix_checksums(self, calc_data = True):
        self.header["data_length"] = self.data_length
        crc32 = self.crc32(self.get_header_length(), self.get_header_length() + self.data_length - 4)
        self.write_header()
        self.write32(-4, crc32, os.SEEK_END)

    def get_chip_id(self):
        return 6

    def get_chip_rev(self):
        return 1

    def wrap(self):
        self.write32(-4, 0, os.SEEK_END)
        super().wrap()
        return True

#   Because somebody fucked up implementing a usual crc32 in bootrom, we have
#   to reinvent the wheel. 
    def crc32(self, from_byte, to_byte=-1):
        crc32_std = super().crc32(from_byte, to_byte)
        self.fd.seek(from_byte, os.SEEK_SET)
        if (to_byte == -1):
            to_byte = self.file_size
        
        polynom = 0x04C11DB7
        crc = 0xffffffff
        pos = from_byte
        while (pos < to_byte):
            data = self.read32(pos)
            pos = pos + 4
            crc = crc ^ data
            for j in range(0, 32):
                if (crc & 0x80000000):
                    crc = crc << 1
                    crc = crc ^ polynom
                else:
                    crc = crc << 1
                crc =  crc & 0xffffffff                     
        crc =  crc ^ 0xffffffff
        return crc



from rumboot.classLoader import classLoader

class ChipDb(classLoader):
    def __getitem__(self, key):
        ret = super().__getitem__(key)
        if ret:
            return ret
        for name, c in self.classes.items():
            if c.chip_id == int(key):
                return c
        return None

    def load(self, path):
        super().load(path)
        for key,value in self.classes.items():
            value.name = key


#!/usr/bin/env python3
#
# Copyright (c) 2018 Andrew Andrianov <andrew@ncrmnt.org>
#

import rumboot
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = "rumboot-tools",
    version = rumboot.__version__,
    packages=find_packages(),
    install_requires = [
        "pyserial", 
        "pyft232", 
        "parse", 
        "xmodem", 
        "tqdm", 
        "pyyaml", 
        "gdbgui", 
        "hexdump",
        "netifaces", 
        "netaddr",
        "getmac",
        "pyusb>=1.0.0",
        "heatshrink2==0.10.0"
        ],

    extras_require = {
    },
    entry_points = {
        "console_scripts": [
            "rumboot-packimage = rumboot_packimage.frontend:cli",
            "rumboot-xrun = rumboot_xrun.frontend:cli",
            "rumboot-xflash = rumboot_xflash.frontend:cli",
            "rumboot-daemon = rumboot_daemon.frontend:cli",
            "rumboot-combine = rumboot_combine.frontend:cli",
            "rumboot-flashrom = rumboot_flashrom.frontend:cli",
            "rumboot-gdb = rumboot_gdb.frontend:cli",
        ],
    },
    include_package_data=True,
    description = "RumBoot ROM Loader Tools",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    keywords = "rumboot image",
    url = "http://github.com/RC-MODULE/rumboot-tools",
    author = "Andrew Andrianov",
    author_email = "andrew@ncrmnt.org",
    license = "GPL",
    platforms = "any",
    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development",
    ],
)



import inspect
import pkgutil

class classLoader():
    iteratable = []

    def __init__(self, objectpath):
        self.opath = objectpath
        self.classes = dict()
        self.load(objectpath)

    def __iter__(self):
        iteratable = []
        for key,value in self.classes.items():
            self.iteratable.append(value)
        return self

    def __next__(self):
        try:
            ret = self.iteratable.pop()
        except:
            raise StopIteration      
        return ret

    def __getitem__(self, key):
        if key in self.classes:
            return self.classes[key]
        for key,value in self.classes.items():
            if type(value).__name__ == key:
                return c            

    def load(self, path):
        p = __import__(path, fromlist=['object'])
        for mod in pkgutil.iter_modules(p.__path__):
            nm = p.__name__ + "." + mod.name
            __import__(nm)

        for name, obj in inspect.getmembers(p):
            if inspect.ismodule(obj):
                p = __import__(obj.__name__, fromlist=['object'])
                for name, cl in inspect.getmembers(p):
                    if inspect.isclass(cl):
                        self.classes[name] = cl



import xml.dom.minidom 
import os

class Report:

    __config = None
    __root = None
    __testsuites = None
    __testsuite = None
    __tests = 0
    __fauilures = 0

    def __init__(self, config):
        self.__config = config
        self.__root = xml.dom.minidom.Document()
        self.__testsuites = self.__root.createElement("testsuites") 
        self.__root.appendChild(self.__testsuites)
        self.__testsuite = self.__root.createElement("testsuite")
        self.__testsuite.setAttribute("name", self.__config["name"])
        self.__testsuites.appendChild(self.__testsuite)

    def add_test_result(self, test, passed):
        self.__tests += 1
        testcase = self.__root.createElement("testcase")
        testcase.setAttribute("name", test)
        if not passed:
            self.__fauilures += 1
            failure = self.__root.createElement("failure")
            failure.setAttribute("message", "Failed")
            testcase.appendChild(failure)
        self.__testsuite.appendChild(testcase)

    def flush(self):
        self.__testsuite.setAttribute("tests", str(self.__tests))
        self.__testsuite.setAttribute("errors", "0")
        self.__testsuite.setAttribute("failures", str(self.__fauilures))

        file_path = os.path.join(self.__config["reports_dir"], self.__config["name"] + ".xml")
        with open(file_path, "w") as file:
            self.__root.writexml(file, addindent="  ", newl="\n", encoding="UTF-8")



import struct
import socket
from getmac import get_mac_address
import time
# sudo arp -s 192.168.144.9 0:0:5e:0:0:0

class edcl_packet:
    offset  = 0
    control = 0
    address = 0

    seq = 0
    flag = 0
    plen = 0

    payload = b''
    FORMAT = "=HII"
    def __init__(self, seq=0):
        self.seq = seq

    def read(self, address, len):
        self.plen = len
        self.address = address
        self.flag = 0

    def write(self, address, data):
        self.payload = data
        self.plen = len(data)
        self.address = address
        self.flag = 1

    def update(self):
        self.control = socket.htonl((self.seq & 0x3FFF) << 18 | ((not not self.flag) & 1) << 17 | (self.plen & 0x3FF) << 7)

    def e_seq(self):
            return socket.ntohl(self.control) >> 18

    def e_rwnak(self):
            return (socket.ntohl(self.control) >> 17) & 1

    def e_len(self):
            return (socket.ntohl(self.control) >> 7) & 0x3FF

    def serialize(self):
        self.update()
        #Make sure we're 4-byte aligned
        self.offset  = self.address & 0x3
        self.address = self.address & ~0x3
        ret = struct.pack(self.FORMAT, self.offset, self.control, socket.htonl(self.address)) + self.payload
        return ret

    def deserialize(self,data):
        self.offset, self.control, self.address = struct.unpack_from(self.FORMAT, data)
        start = struct.calcsize(self.FORMAT)
        self.payload = data[start:]
        if (self.e_len() != len(self.payload)):
            raise Exception("edcl_packet: Whoops, got length %d, expected %d" % (len(self.payload), self.e_len()))
        self.len = self.e_len()
        self.seq = self.e_seq()
        
    def add_data(self, data):
        self.payload += data
        self.plen = len(self.payload)

    def dump(self):
        print("addr: ", self.address)
        print("control: ", self.control)
        print("offset: ", self.offset)
        print("len: ", self.plen)
        print("seq: ", self.seq)
        print("RWNAK: ", self.e_rwnak())
        print("------")
    def data(self):
        return self.payload

class edcl():
    sock = None
    local_port = 0x8088
    remote_port = 0x9099
    remote_ip = "192.168.0.1"
    maxpayload = 456
    swap_endian = False
    seq = 0

    stats = {
        "noreply" : 0,
        "lostack" : 0,
        "rwnak": 0,
        "badseq" : 0
    }

    def __init__(self):        
        rc = socket.socket(socket.AF_INET, # Internet
                             socket.SOCK_DGRAM) # UDP
        rc.bind(("0.0.0.0", 0x8088))
        self.sock = rc

    def set_max_payload(self, p):
        self.maxpayload = p

    def xfer(self, packet, checkrwnak=True, retries=32):
        rcv = edcl_packet()
        noreply = False
        while packet.plen % 4:
            packet.add_data(b'\x00')

        for i in range(0,retries):
            packet.seq = self.seq
            sentlen = 0
            try:
                self.sock.sendto(packet.serialize(), (self.remote_ip, self.remote_port))
                data, addr = self.sock.recvfrom(struct.calcsize(edcl_packet.FORMAT) + self.maxpayload)
                if addr[0] != self.remote_ip:
                    continue
                rcv.deserialize(data)
            except Exception as e:
                self.stats["noreply"] +=1
                noreply = True
                continue

            if not checkrwnak:
                return rcv

            if rcv.e_rwnak():
                self.stats["rwnak"] +=1
                if noreply and rcv.e_seq() == packet.e_seq() + 1:
                    noreply = False
                    self.seq = rcv.e_seq()
                    self.stats["lostack"] +=1
                    continue

                if rcv.e_seq() != packet.e_seq():
                    self.stats["badseq"] +=1
                    print("edcl: Sync lost. got %d expected %d" % (rcv.e_seq(), packet.e_seq()))
                    self.seq = rcv.e_seq()
                    continue

            if rcv.e_seq() == packet.e_seq():
                self.seq = self.seq + 1                
                return rcv
        raise Exception("EDCL Transfer failed")

    def _read_raw_(self, address, len):
        tx = edcl_packet()
        tx.read(address, len)
        rx = self.xfer(tx)
        return rx.payload

    def _write_raw_(self, address, data):
        tx = edcl_packet()
        tx.write(address, data)
        rx = self.xfer(tx)
        return (rx.e_rwnak() == 0)

    def _read_raw(self, address, len):
        if self.swap_endian and (address % 4 or len % 4):
            raise Exception("edcl: Requested misaligned operation for slave that needs endian swap")
        #TODO: Swap endian here
        return self._read_raw_(address, len)

    def _write_raw(self, address, data):
        if self.swap_endian and (address % 4 or len % 4):
            raise Exception("edcl: Requested misaligned operation for slave that needs endian swap")
        #TODO: Swap endian here
        return self._write_raw_(address, data)

    def read32(self, address):
        data = _read_raw(self, address, 4)
        return struct.unpack_from("=I",data)

    def write32(self, address, data):
        data = struct.pack("=I",  data)
        self._write_raw(address, data)

    def read(self, address, len, callback = None):
        ret = b''
        for l in range (0, len, self.maxpayload):
            if (len - l > self.maxpayload):
                toread = self.maxpayload
            else:
                toread = len - l
            ret = ret + self._read_raw(address + l, toread)
            if callback:
                callback(len, l + toread, toread)
        return ret

    def write(self, address, data, callback = None):
        length = len(data)
        for l in range (0, length, self.maxpayload):
            if (length - l > self.maxpayload):
                towrite = self.maxpayload
            else:
                towrite = length - l

            self._write_raw(address + l, data[l:l+towrite])
            if callback:
                callback(len, l + towrite, towrite)


    def send_from_file(self, address, fl, callback = None, offset = 0, length = -1):
        if type(fl) == str:
            fl = open(fl, "rb")

        if length == -1:
            fl.seek(0, 2)
            length = fl.tell() - offset

        fl.seek(offset)
        chunksize = self.maxpayload * 10
        def wrapcb(total, position, lastwrite):
            if callback:
                callback(length, position, lastwrite)

        while True:
            toread = chunksize
            if toread > length:
                toread = length
            chunk = fl.read(length)
            if len(chunk) == 0:
                break
            self.write(address, chunk, wrapcb)
            address = address + chunksize
            length -= toread
            if length == 0:
                break
#        print(self.stats)
        
    def recv_to_file(self, address, length, fl, callback = None):
        if type(fl) == str:
            fl = open(fl, "wb+")
            needclose = True
        else:
            needclose = False
        chunksize = self.maxpayload * 10
        def wrapcb(total, position, lastwrite):
            if callback:
                callback(length, position, lastwrite)
        rd = 0
        while rd < length:
            chunk = self.read(address + rd, chunksize, wrapcb)
            fl.write(chunk)
            rd = rd + chunksize
        if needclose:
            fl.close()
#        print(self.stats)

    def reconnect(self):
        return self.connect(remote_ip=self.remote_ip, remote_port=self.remote_port)

    def connect(self, remote_ip="192.168.144.9", remote_port=0x9099):
        self.sock.settimeout(0.6) # the timeout should include ARP request timeout overwise the requests will be accumulated and next requests will be delayed
        self.remote_ip = remote_ip
        self.remote_port = remote_port
        tx = edcl_packet()

        # Get rid of any residual data we may get
        # Works around some rare corner-cases
        try: 
            self.sock.recvfrom(4096)
        except:
            pass

        try:
            rx = self.xfer(tx, False, 1)
            if rx.e_rwnak():
                self.seq = rx.e_seq()
            else:
                self.seq = rx.e_seq() + 1
            self.seq = rx.e_seq() + 1
            self.sock.settimeout(0.1)
            for shit in self.stats:
                self.stats[shit]=0
            return (tx.address == rx.address) and (tx.e_len() == rx.e_len())
        except Exception as e:
            return False

